"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

flows.py
~~~~~~~~~~~~~~~

Plot energy flows data.
"""

import pandas as pd

from calliope.postprocess.plotting.util import break_name, get_range


def _line(
    locs_coordinates,
    transmission_type,
    to_location,
    from_location,
    carrier,
    tech,
    prod,
    scale_factor,
    techs_colors,
    is_initial_timestep,
    add_legend,
    name,
):
    # e.g. "Region1->Region2: 256.54 by gas_transmission (gas)"
    hover_info = "%s->%s: %.2f by %s (%s)" % (
        from_location,
        to_location,
        prod,
        transmission_type,
        carrier,
    )

    line = dict(
        visible=False,
        mode="lines",
        hoverinfo="text",
        text="",
        line=dict(width=prod * scale_factor + 1, color=techs_colors[transmission_type]),
        legendgroup=transmission_type,
        opacity=0.6,
        showlegend=False,
    )

    line_legend = dict(
        visible=False,
        mode="lines",
        hoverinfo="text",
        text="",
        line=dict(width=10, color=techs_colors[transmission_type]),
        legendgroup=transmission_type,
        name=break_name(name, 18),
        opacity=0.6,
    )

    line_info_marker = dict(
        visible=False,
        mode="markers",
        hoverinfo="text",
        text=hover_info,
        marker=dict(symbol="square", opacity=0, color=techs_colors[transmission_type]),
        legendgroup=transmission_type,
        name=tech,
        showlegend=False,
    )

    if set(locs_coordinates.index) == set(["x", "y"]):
        h_coord, v_coord = "x", "y"
    elif set(locs_coordinates.index) == set(["lon", "lat"]):
        h_coord, v_coord = "lon", "lat"

    line[h_coord] = [
        locs_coordinates[from_location][h_coord],
        locs_coordinates[to_location][h_coord],
    ]
    line[v_coord] = [
        locs_coordinates[from_location][v_coord],
        locs_coordinates[to_location][v_coord],
    ]
    line_legend[h_coord] = [None]
    line_legend[v_coord] = [None]
    line_info_marker[h_coord] = [
        (1 / 2)
        * (
            locs_coordinates[from_location][h_coord]
            + locs_coordinates[to_location][h_coord]
        )
    ]
    line_info_marker[v_coord] = [
        (1 / 2)
        * (
            locs_coordinates[from_location][v_coord]
            + locs_coordinates[to_location][v_coord]
        )
    ]

    if is_initial_timestep:
        # plot only the first timestep data when the chart is initialized
        line["visible"] = True
        line_legend["visible"] = True
        line_info_marker["visible"] = True
    if add_legend:
        return [line, line_legend, line_info_marker]
    else:
        return [line, line_info_marker]


def _marker(
    locs_coordinates,
    location,
    carrier,
    tech,
    prod,
    scale_factor,
    techs_colors,
    is_initial_timestep,
    add_legend,
    name,
):
    # Example: "Region1: 3552.65 of pipe_import (gas)"
    hover_info = "%s: %.2f of %s (%s)" % (location, prod, tech, carrier)

    marker_dict = dict(
        visible=False,
        hoverinfo="text",
        text=hover_info,
        mode="markers",
        marker=dict(
            symbol="circle-dot",
            opacity=0.6,
            size=prod * scale_factor + 1,
            color=techs_colors[tech],
        ),
        legendgroup=tech,
        showlegend=False,
    )

    marker_legend = dict(
        visible=False,
        hoverinfo="text",
        text=hover_info,
        mode="markers",
        marker=dict(
            symbol="circle-dot",
            opacity=0.6,
            size=10,
            color=techs_colors[tech],
        ),
        legendgroup=tech,
        name=break_name(name, 18),
    )

    if set(locs_coordinates.index) == set(["x", "y"]):
        h_coord, v_coord = "x", "y"
    elif set(locs_coordinates.index) == set(["lon", "lat"]):
        h_coord, v_coord = "lon", "lat"

    marker_dict[h_coord] = [locs_coordinates[location][h_coord]]
    marker_dict[v_coord] = [locs_coordinates[location][v_coord]]
    marker_legend[h_coord] = [None]
    marker_legend[v_coord] = [None]

    if is_initial_timestep:
        # plot only the first timestep data when the chart is initialized
        marker_dict["visible"] = True
        marker_legend["visible"] = True
    if add_legend:
        return [marker_dict, marker_legend]
    else:
        return [marker_dict]


def _production_data(model, timesteps, timestep):
    """
    returns a list of dicts, each dict is a plotly marker (location production)
    or line (transmission) on the map.
    """
    locs_coordinates = model._model_data.loc_coordinates.to_pandas()
    locs_techs_carriers_production = model.get_formatted_array("carrier_prod")
    techs_colors = model._model_data.colors.to_pandas()
    scale_factor = 100 / abs(
        model.results.carrier_prod.values.max()
        - model.results.carrier_prod.values.min()
    )
    tech_names = set(model._model_data.techs.values)

    production_data = []
    # we iterate through each dimension for one timestep in order to
    # add the different production sources of the location and line
    # transmissions toward it
    # Complexity: O(len(carriers)*len(techs)*len(locations)). Considering
    # len(carriers) is supposed to be small (<10), we have a large margin in
    # terms of techs and locations numbers (len(techs)*len(locations) <= 10^9,
    # equivalent to one second processing time for python)
    links = []
    # list of sets { tech, from_location, to_location }
    links_data = []
    # links associated data, like prod, carrier, transmission_type
    # [ [prod, carrier, transmission_type], [] ..]
    for location in locs_techs_carriers_production.locs.values:

        for carrier in locs_techs_carriers_production.carriers.values:
            techs_production = locs_techs_carriers_production.sel(
                carriers=carrier, locs=location
            ).to_pandas()
            for tech, prod in techs_production.loc[:, timestep].iteritems():
                if prod and prod > 0:
                    # if some energy is at stake

                    tech_name = tech.split(":")[0]
                    if tech_name in tech_names:
                        add_legend = True
                        tech_names.discard(tech_name)
                        name = model._model_data.names.loc[tech_name].item()
                    else:
                        # only add legend information once for a tech
                        add_legend = False
                        name = ""

                    if len(tech.split(":")) > 1:
                        # "transmission_type:location"
                        # if it gets energy from another location
                        [transmission_type, from_location] = tech.split(":")
                        links.append({tech_name, from_location, location})
                        links_data.append(
                            {
                                "transmission_type": transmission_type,
                                "from_location": from_location,
                                "to_location": location,
                                "prod": prod,
                                "carrier": carrier,
                                "tech": tech,
                                "add_legend": add_legend,
                                "name": name,
                            }
                        )
                    else:
                        # if the energy comes from this location
                        production_data.extend(
                            _marker(
                                locs_coordinates,
                                location,
                                carrier,
                                tech,
                                prod,
                                scale_factor,
                                techs_colors,
                                timestep == timesteps[0],
                                add_legend,
                                name,
                            )
                        )

    def merge(first_link, second_link):
        if first_link["prod"] > second_link["prod"]:
            first_link["prod"] -= second_link["prod"]
            return first_link
        elif first_link["prod"] < second_link["prod"]:
            second_link["prod"] -= first_link["prod"]
            return second_link
        else:
            # the two transmission links are equal,
            # thus, no representation of it is return
            return {}

    # merge the links data
    links_merged = []
    while len(links) > 0:
        data = links_data[0]
        # we check if there is a transmission
        # link in the opposite direction and merge if so
        j = 1
        while j < len(links):
            if links[j] == links[0]:
                data = merge(links_data[0], links_data[j])
                links.remove(links[j])
                links_data.remove(links_data[j])
                j -= 1
            j += 1
        links_merged.append(data)
        links.remove(links[0])
        links_data.remove(links_data[0])

    # add merged links to production_data
    for link in links_merged:
        if link:
            params_list = [
                locs_coordinates,
                link["transmission_type"],
                link["to_location"],
                link["from_location"],
                link["carrier"],
                link["tech"],
                link["prod"],
                scale_factor,
                techs_colors,
                timestep == timesteps[0],
                link["add_legend"],
                link["name"],
            ]
            production_data.extend(_line(*params_list))

    return production_data


def plot_flows(model, timestep_cycle=1, timestep_index_subset=[], **kwargs):
    """
    Parameters
    ----------
    timestep_cycle : int, optional
        Shows one of every timestep_cycle timesteps. Default is 1 (all timesteps
        are shown).
    timestep_index_subset : list of int, optional
        Only the timesteps between those two indexes are shown. Default is []
        (all timesteps are shown).
    """

    if len(timestep_index_subset) == 2:
        timestep_start = timestep_index_subset[0]
        timestep_end = timestep_index_subset[1]
    else:
        timestep_start, timestep_end = 0, len(model._model_data.timesteps.values)

    try:
        locs_coordinates = model._model_data.loc_coordinates
    except AttributeError:
        raise ValueError(
            "Model does not define location coordinates "
            "- no energy flow plotting possible."
        )

    timesteps = model._model_data.timesteps.values[
        timestep_start:timestep_end:timestep_cycle
    ]  # slicing the desired timesteps
    timeseries_dateformat = model.model_config["timeseries_dateformat"]

    steps_length = []
    data = []
    for timestep in timesteps:
        data_by_timestep = _production_data(model, timesteps, timestep)
        steps_length.append(len(data_by_timestep))
        data.extend(data_by_timestep)

    steps = []
    for i, timestep in enumerate(timesteps):
        step = dict(
            # active="label of first show timestep data",
            method="restyle",
            args=["visible", [False] * len(data)],
            label=pd.to_datetime(timestep).strftime(timeseries_dateformat),
        )
        i_start = sum(steps_length[:i])  # visible start index
        i_end = i_start + steps_length[i]  # visible end index
        step["args"][1][i_start:i_end] = [True] * steps_length[i]
        # we set visible to True for all the points of one timestep
        steps.append(step)

    sliders = [
        dict(
            # active="start sliding",True
            currentvalue=dict(
                visible=True,
                prefix="Timestep: ",
            ),
            pad={"t": 50},
            activebgcolor="black",
            bgcolor="grey",
            steps=steps,
        )
    ]

    # define the map general layout here
    layout = dict(
        title="Energy Flow",
        showlegend=True,
        width=900,
        height=700,
        hovermode="closest",
        sliders=sliders,
        margin={"autoexpand": False, "b": 150, "r": 180},
    )

    # change the range of the plot whether its x,y or lat,lon coords
    if sorted(locs_coordinates.coordinates.values) == ["x", "y"]:
        layout["xaxis"] = dict(range=get_range(locs_coordinates, "x", 0.2))
        layout["yaxis"] = dict(range=get_range(locs_coordinates, "y", 0.2))
        for trace in data:
            trace["type"] = "scatter"
    elif sorted(locs_coordinates.coordinates.values) == ["lat", "lon"]:
        layout["geo"] = dict(
            scope="world",
            showland=True,
            showcountries=True,
            showsubunits=True,
            showocean=True,
            oceancolor="#aec6cf",
            subunitcolor="blue",
            countrycolor="green",
            lonaxis=dict(range=get_range(locs_coordinates, "lon", 0.2)),
            lataxis=dict(range=get_range(locs_coordinates, "lat", 0.2)),
            countrywidth=0.5,
            subunitwidth=0.5,
            landcolor="rgb(255,255,255)",
        )
        for trace in data:
            trace["type"] = "scattergeo"

    return data, layout



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

preprocess_model.py
~~~~~~~~~~~~~~~~~~~

Preprocessing of model and run configuration into a unified model_run
AttrDict, and building of associated debug information.

"""

import os
import logging
import itertools
import warnings

import pandas as pd

import calliope
from calliope import exceptions
from calliope.core.attrdict import AttrDict
from calliope.core.util.tools import relative_path
from calliope.preprocess import locations, sets, checks, constraint_sets, util

logger = logging.getLogger(__name__)


# Output of: sns.color_palette('cubehelix', 10).as_hex()
_DEFAULT_PALETTE = [
    "#19122b",
    "#17344c",
    "#185b48",
    "#3c7632",
    "#7e7a36",
    "#bc7967",
    "#d486af",
    "#caa9e7",
    "#c2d2f3",
    "#d6f0ef",
]


def model_run_from_yaml(
    model_file, timeseries_dataframes=None, scenario=None, override_dict=None
):
    """
    Generate processed ModelRun configuration from a
    YAML model configuration file.

    Parameters
    ----------
    model_file : str
        Path to YAML file with model configuration.
    timeseries_dataframes : dict, optional
        Dictionary of timeseries dataframes. The keys are strings
        corresponding to the dataframe names given in the yaml files and
        the values are dataframes with time series data.
    scenario : str, optional
        Name of scenario to apply. Can either be a named scenario, or a
        comma-separated list of individual overrides to be combined
        ad-hoc, e.g. 'my_scenario_name' or 'override1,override2'.
    override_dict : dict or AttrDict, optional

    """
    config = AttrDict.from_yaml(model_file)
    config.config_path = model_file

    config_with_overrides, debug_comments, overrides, scenario = apply_overrides(
        config, scenario=scenario, override_dict=override_dict
    )

    return generate_model_run(
        config_with_overrides,
        timeseries_dataframes,
        debug_comments,
        overrides,
        scenario,
    )


def model_run_from_dict(
    config_dict, timeseries_dataframes=None, scenario=None, override_dict=None
):
    """
    Generate processed ModelRun configuration from a
    model configuration dictionary.

    Parameters
    ----------
    config_dict : dict or AttrDict
    timeseries_dataframes : dict, optional
        Dictionary of timeseries dataframes. The keys are strings
        corresponding to the dataframe names given in the yaml files and
        the values are dataframes with time series data.
    scenario : str, optional
        Name of scenario to apply. Can either be a named scenario, or a
        comma-separated list of individual overrides to be combined
        ad-hoc, e.g. 'my_scenario_name' or 'override1,override2'.
    override_dict : dict or AttrDict, optional

    """
    if not isinstance(config_dict, AttrDict):
        config = AttrDict(config_dict)
    else:
        config = config_dict
    config.config_path = None

    config_with_overrides, debug_comments, overrides, scenario = apply_overrides(
        config, scenario=scenario, override_dict=override_dict
    )

    return generate_model_run(
        config_with_overrides,
        timeseries_dataframes,
        debug_comments,
        overrides,
        scenario,
    )


def combine_overrides(config_model, overrides):
    override_dict = AttrDict()
    for override in overrides:
        try:
            yaml_string = config_model.overrides[override].to_yaml()
            override_with_imports = AttrDict.from_yaml_string(yaml_string)
        except KeyError:
            raise exceptions.ModelError(
                "Override `{}` is not defined.".format(override)
            )
        try:
            override_dict.union(override_with_imports, allow_override=False)
        except KeyError as e:
            raise exceptions.ModelError(
                str(e)[1:-1] + ". Already specified but defined again in "
                "override `{}`.".format(override)
            )

    return override_dict


def apply_overrides(config, scenario=None, override_dict=None):
    """
    Generate processed Model configuration, applying any scenarios overrides.

    Parameters
    ----------
    config : AttrDict
        a model configuration AttrDict
    scenario : str, optional
    override_dict : str or dict or AttrDict, optional
        If a YAML string, converted to AttrDict

    """
    debug_comments = AttrDict()

    config_model = AttrDict.from_yaml(
        os.path.join(os.path.dirname(calliope.__file__), "config", "defaults.yaml")
    )

    # Interpret timeseries_data_path as relative
    if "timeseries_data_path" in config.model:
        config.model.timeseries_data_path = relative_path(
            config.config_path, config.model.timeseries_data_path
        )

    # FutureWarning: check if config includes an explicit objective cost class.
    # Added in 0.6.4-dev, to be removed in v0.7.0-dev.
    has_explicit_cost_class = isinstance(
        config.get_key("run.objective_options.cost_class", None), dict
    )

    # The input files are allowed to override other model defaults
    config_model.union(config, allow_override=True)

    # First pass of applying override dict before applying scenarios,
    # so that can override scenario definitions by override_dict
    if override_dict:
        if isinstance(override_dict, str):
            override_dict = AttrDict.from_yaml_string(override_dict)
        elif not isinstance(override_dict, AttrDict):
            override_dict = AttrDict(override_dict)

        warning_messages = checks.check_overrides(config_model, override_dict)
        exceptions.print_warnings_and_raise_errors(warnings=warning_messages)

        # FutureWarning: If config does not include an explicit objective cost class, check override dict.
        # Added in 0.6.4-dev, to be removed in v0.7.0-dev.
        if has_explicit_cost_class is False:
            has_explicit_cost_class = isinstance(
                override_dict.get_key("run.objective_options.cost_class", None), dict
            )

        config_model.union(override_dict, allow_override=True, allow_replacement=True)

    if scenario:
        scenario_overrides = load_overrides_from_scenario(config_model, scenario)
        if not all(i in config_model.get("overrides", {}) for i in scenario_overrides):
            raise exceptions.ModelError(
                "Scenario definition must be a list of override or other scenario names."
            )
        else:
            logger.info(
                "Applying the following overrides from scenario definition: {} ".format(
                    scenario_overrides
                )
            )
        overrides_from_scenario = combine_overrides(config_model, scenario_overrides)

        warning_messages = checks.check_overrides(config_model, overrides_from_scenario)
        exceptions.print_warnings_and_raise_errors(warnings=warning_messages)

        # FutureWarning: If config nor override_dict include an explicit objective cost class, check scenario dict.
        # Added in 0.6.4-dev, to be removed in v0.7.0-dev
        if has_explicit_cost_class is False:
            has_explicit_cost_class = isinstance(
                overrides_from_scenario.get_key(
                    "run.objective_options.cost_class", None
                ),
                dict,
            )

        config_model.union(
            overrides_from_scenario, allow_override=True, allow_replacement=True
        )
        for k, v in overrides_from_scenario.as_dict_flat().items():
            debug_comments.set_key("{}".format(k), "Applied from override")
    else:
        scenario_overrides = []

    # Second pass of applying override dict after applying scenarios,
    # so that scenario-based overrides are overridden by override_dict!
    if override_dict:
        config_model.union(override_dict, allow_override=True, allow_replacement=True)
        for k, v in override_dict.as_dict_flat().items():
            debug_comments.set_key(
                "{}".format(k), "Overridden via override dictionary."
            )

    # FutureWarning: raise cost class warning here.
    # Warning that there will be no default cost class in 0.7.0 #
    # Added in 0.6.4-dev, to be removed in v0.7.0-dev
    if has_explicit_cost_class is False:
        warnings.warn(
            "There will be no default cost class for the objective function in "
            'v0.7.0 (currently "monetary" with a weight of 1). '
            "Explicitly specify the cost class(es) you would like to use "
            'under `run.objective_options.cost_class`. E.g. `{"monetary": 1}` to '
            "replicate the current default.",
            FutureWarning,
        )

    # Drop default locations, links, and techs
    config_model.del_key("techs.default_tech")
    config_model.del_key("locations.default_location")
    config_model.del_key("links.default_location_from,default_location_to")
    config_model.del_key("group_constraints.default_group")

    return config_model, debug_comments, scenario_overrides, scenario


def get_parents(tech_id, model_config):
    """
    Returns the full inheritance tree from which ``tech`` descends,
    ending with its base technology group.

    To get the base technology group,
    use ``get_parents(...)[-1]``.

    Parameters
    ----------
    tech : str
    model_config : AttrDict

    """

    tech = model_config.techs[tech_id].essentials.parent
    parents = [tech]

    while True:
        tech = model_config.tech_groups[tech].essentials.parent
        if tech is None:
            break  # We have reached the top of the chain
        parents.append(tech)
    return parents


def process_techs(config_model):

    default_palette_cycler = itertools.cycle(range(len(_DEFAULT_PALETTE)))

    result = AttrDict()
    errors = []
    debug_comments = AttrDict()

    for tech_id, tech_config in config_model.techs.items():

        # If a tech specifies ``exists: false``, we skip it entirely
        if not tech_config.get("exists", True):
            continue

        tech_result = AttrDict()

        # Add inheritance chain
        tech_result.inheritance = get_parents(tech_id, config_model)

        # CHECK: A tech's parent must lead to one of the built-in tech_groups
        builtin_tech_groups = checks.DEFAULTS.tech_groups.keys()
        if tech_result.inheritance[-1] not in builtin_tech_groups:
            errors.append(
                "tech {} must inherit from a built-in tech group".format(tech_id)
            )

        # Process inheritance
        tech_result.essentials = AttrDict()
        tech_result.constraints = AttrDict()
        for parent in reversed(tech_result.inheritance):
            # Does the parent group have model-wide settings?
            parent_essentials = config_model.tech_groups[parent].essentials
            parent_systemwide_constraints = util.get_systemwide_constraints(
                config_model.tech_groups[parent]
            )
            for k in parent_essentials.as_dict_flat():
                debug_comments.set_key(
                    "{}.essentials.{}".format(tech_id, k),
                    "From parent tech_group `{}`".format(parent),
                )
            tech_result.essentials.union(parent_essentials, allow_override=True)
            tech_result.constraints.union(
                parent_systemwide_constraints, allow_override=True
            )

        # Add this tech's essentials and constraints, overwriting any essentials from parents
        tech_result.essentials.union(tech_config.essentials, allow_override=True)
        tech_result.constraints.union(
            util.get_systemwide_constraints(tech_config), allow_override=True
        )

        # Add allowed_constraints and required_constraints from base tech
        keys_to_add = [
            "required_constraints",
            "allowed_constraints",
            "allowed_group_constraints",
            "allowed_costs",
        ]
        for k in keys_to_add:
            tech_result[k] = config_model.tech_groups[tech_result.inheritance[-1]].get(
                k, []
            )

        # CHECK: If necessary, populate carrier_in and carrier_out in essentials, but
        # also break on missing carrier data
        if "carrier_in" not in tech_result.essentials:
            if tech_result.inheritance[-1] in ["supply", "supply_plus"]:
                tech_result.essentials.carrier_in = "resource"
            elif tech_result.inheritance[-1] in ["demand", "transmission", "storage"]:
                try:
                    tech_result.essentials.carrier_in = tech_result.essentials.carrier
                    debug_comments.set_key(
                        "{}.essentials.carrier_in".format(tech_id),
                        "Set from essentials.carrier",
                    )
                except KeyError:
                    errors.append(
                        "`carrier` or `carrier_in` must be "
                        "defined for {}".format(tech_id)
                    )
            else:
                errors.append("`carrier_in` must be defined for {}".format(tech_id))

        if "carrier_out" not in tech_result.essentials:
            if tech_result.inheritance[-1] == "demand":
                tech_result.essentials.carrier_out = "resource"
            elif tech_result.inheritance[-1] in [
                "supply",
                "supply_plus",
                "transmission",
                "storage",
            ]:
                try:
                    tech_result.essentials.carrier_out = tech_result.essentials.carrier
                except KeyError:
                    errors.append(
                        "`carrier` or `carrier_out` must be "
                        "defined for {}".format(tech_id)
                    )
            else:
                errors.append("`carrier_out` must be defined for {}".format(tech_id))
        # Deal with primary carrier in/out for conversion_plus techs
        if tech_result.inheritance[-1] == "conversion_plus":
            for direction in ["_in", "_out"]:
                carriers = set(
                    util.flatten_list(
                        [
                            v
                            for k, v in tech_result.essentials.items()
                            if k.startswith("carrier" + direction)
                        ]
                    )
                )
                primary_carrier = tech_result.essentials.get(
                    "primary_carrier" + direction, None
                )
                if primary_carrier is None and len(carriers) == 1:
                    tech_result.essentials[
                        "primary_carrier" + direction
                    ] = carriers.pop()
                elif primary_carrier is None and len(carriers) > 1:
                    errors.append(
                        "Primary_carrier{0} must be assigned for tech `{1}` as "
                        "there are multiple carriers{0}".format(direction, tech_id)
                    )
                elif primary_carrier not in carriers:
                    errors.append(
                        "Primary_carrier{0} `{1}` not one of the available carriers"
                        "{0} for `{2}`".format(direction, primary_carrier, tech_id)
                    )

        # If necessary, pick a color for the tech, cycling through
        # the hardcoded default palette
        if not tech_result.essentials.get_key("color", None):
            color = _DEFAULT_PALETTE[next(default_palette_cycler)]
            tech_result.essentials.color = color
            debug_comments.set_key(
                "{}.essentials.color".format(tech_id), "From Calliope default palette"
            )
        result[tech_id] = tech_result

    return result, debug_comments, errors


def load_overrides_from_scenario(config_model, scenario):
    def _get_overrides(scenario_name):
        _overrides = config_model.get_key(f"scenarios.{scenario_name}", [scenario_name])
        if isinstance(_overrides, list):
            return _overrides
        else:
            return [_overrides]

    if scenario in config_model.get("scenarios", {}).keys():
        if "," in scenario:
            warnings.warn(
                f"Scenario name `{scenario}` includes commas that won't be parsed as a list of overrides."
            )
        logger.info("Loading overrides from scenario: {} ".format(scenario))
        scenario_list = _get_overrides(scenario)
    else:
        scenario_list = scenario.split(",")
    scenario_overrides = set()
    for override in scenario_list:
        if isinstance(override, dict):
            raise exceptions.ModelError(
                "Scenario definition must be a list of override or other scenario names."
            )
        if override in config_model.get("scenarios", {}).keys():
            scenario_overrides.update(
                load_overrides_from_scenario(config_model, override)
            )
        else:
            scenario_overrides.add(override)

    return list(scenario_overrides)


def process_tech_groups(config_model, techs):
    tech_groups = AttrDict()
    for group in config_model.tech_groups.keys():
        members = set(k for k, v in techs.items() if group in v.inheritance)
        tech_groups[group] = sorted(list(members))
    return tech_groups


def load_timeseries_from_file(config_model, tskey):
    file_path = os.path.join(config_model.model.timeseries_data_path, tskey)
    df = pd.read_csv(file_path, index_col=0)
    return df


def check_timeseries_dataframes(timeseries_dataframes):
    """
    Timeseries dataframes should be dict of pandas DataFrames.
    """
    if not isinstance(timeseries_dataframes, dict) or not all(
        [
            isinstance(timeseries_dataframes[i], pd.DataFrame)
            for i in timeseries_dataframes
        ]
    ):
        raise exceptions.ModelError(
            "Error in loading timeseries data from dataframes. "
            "`timeseries_dataframes` must be dict of pandas DataFrames."
        )


def load_timeseries_from_dataframe(timeseries_dataframes, tskey):

    # If `df=` is called, timeseries_dataframes must be entered
    if timeseries_dataframes is None:
        raise exceptions.ModelError(
            "Error in loading timeseries. Model config specifies df={} but "
            "no timeseries passed as arguments in calliope.Model(...). "
            "Note that, if running from a command line, it is not possible "
            "to read dataframes via `df=...` and you should specify "
            "`file=...` with a CSV file.".format(tskey)
        )

    try:
        df = timeseries_dataframes[tskey]
    except KeyError:
        raise exceptions.ModelError(
            "Error in loading data from dataframe. "
            "Model attempted to load dataframe with key `{}`, "
            "but time series passed as arguments are {}".format(
                tskey, set(timeseries_dataframes.keys())
            )
        )
    if not isinstance(df, pd.DataFrame):
        raise exceptions.ModelError(
            "Error in loading data. Object passed in time series "
            "dictionary under key {} is a {}, not a DataFrame.".format(tskey, type(df))
        )
    return df


def process_timeseries_data(config_model, model_run, timeseries_dataframes):

    if config_model.model.timeseries_data is None:
        timeseries_data = AttrDict()
    else:
        timeseries_data = config_model.model.timeseries_data

    def _parser(x, dtformat):
        return pd.to_datetime(x, format=dtformat, exact=False)

    dtformat = config_model.model["timeseries_dateformat"]

    # Generate set of all files and dataframes we want to load
    location_config = model_run.locations.as_dict_flat()
    model_config = config_model.model.as_dict_flat()

    # Find names of csv files (file=) or dataframes (df=) called in config
    def get_names(datatype, config):
        return set(
            [
                v.split("=")[1].rsplit(":", 1)[0]
                for v in config.values()
                if str(datatype) + "=" in str(v)
            ]
        )

    constraint_filenames = get_names("file", location_config)
    cluster_filenames = get_names("file", model_config)
    constraint_dfnames = get_names("df", location_config)
    cluster_dfnames = get_names("df", model_config)

    # Check if timeseries_dataframes is in the correct format (dict of
    # pandas DataFrames)
    if timeseries_dataframes is not None:
        check_timeseries_dataframes(timeseries_dataframes)

    datetime_min = []
    datetime_max = []

    # Check there is at least one timeseries present
    if (
        len(
            constraint_filenames
            | cluster_filenames
            | constraint_dfnames
            | cluster_dfnames
        )
        == 0
    ):
        raise exceptions.ModelError(
            "There is no timeseries in the model. At least one timeseries is "
            "necessary to run the model."
        )

    # Load each timeseries into timeseries data. tskey is either a filename
    # (called by file=...) or a key in timeseries_dataframes (called by df=...)
    for tskey in (
        constraint_filenames | cluster_filenames | constraint_dfnames | cluster_dfnames
    ):  # Filenames or dict keys
        # If tskey is a CSV path, load the CSV, else load the dataframe
        if tskey in constraint_filenames | cluster_filenames:
            df = load_timeseries_from_file(config_model, tskey)
        elif tskey in constraint_dfnames | cluster_dfnames:
            df = load_timeseries_from_dataframe(timeseries_dataframes, tskey)

        try:
            df.apply(pd.to_numeric)
        except ValueError as e:
            raise exceptions.ModelError(
                "Error in loading data from {}. Ensure all entries are "
                "numeric. Full error: {}".format(tskey, e)
            )
        # Parse the dates, checking for errors specific to this
        try:
            df.index = _parser(df.index, dtformat)
        except ValueError as e:
            raise exceptions.ModelError(
                "Error in parsing dates in timeseries data from {}, "
                "using datetime format `{}`: {}".format(tskey, dtformat, e)
            )
        timeseries_data[tskey] = df

        datetime_min.append(df.index[0].date())
        datetime_max.append(df.index[-1].date())

    # Apply time subsetting, if supplied in model_run
    subset_time_config = config_model.model.subset_time
    if subset_time_config is not None:
        # Test parsing dates first, to make sure they fit our required subset format

        try:
            subset_time = _parser(subset_time_config, "%Y-%m-%d %H:%M:%S")
        except ValueError as e:
            raise exceptions.ModelError(
                "Timeseries subset must be in ISO format (anything up to the  "
                "detail of `%Y-%m-%d %H:%M:%S`.\n User time subset: {}\n "
                "Error caused: {}".format(subset_time_config, e)
            )
        if isinstance(subset_time_config, list) and len(subset_time_config) == 2:
            time_slice = slice(subset_time_config[0], subset_time_config[1])

            # Don't allow slicing outside the range of input data
            if subset_time[0].date() < max(datetime_min) or subset_time[1].date() > min(
                datetime_max
            ):

                raise exceptions.ModelError(
                    "subset time range {} is outside the input data time range "
                    "[{}, {}]".format(
                        subset_time_config,
                        max(datetime_min).strftime("%Y-%m-%d"),
                        min(datetime_max).strftime("%Y-%m-%d"),
                    )
                )
        elif isinstance(subset_time_config, list):
            raise exceptions.ModelError(
                "Invalid subset_time value: {}".format(subset_time_config)
            )
        else:
            time_slice = str(subset_time_config)

        for k in timeseries_data.keys():
            timeseries_data[k] = timeseries_data[k].loc[time_slice, :]
            if timeseries_data[k].empty:
                raise exceptions.ModelError(
                    "The time slice {} creates an empty timeseries array for {}".format(
                        time_slice, k
                    )
                )

    # Ensure all timeseries have the same index
    indices = [
        (tskey, df.index)
        for tskey, df in timeseries_data.items()
        if tskey not in cluster_filenames | cluster_dfnames
    ]
    first_tskey, first_index = indices[0]
    for tskey, idx in indices[1:]:
        if not first_index.equals(idx):
            raise exceptions.ModelError(
                "Time series indices do not match "
                "between {} and {}".format(first_tskey, tskey)
            )

    return timeseries_data, first_index


def generate_model_run(
    config, timeseries_dataframes, debug_comments, applied_overrides, scenario
):
    """
    Returns a processed model_run configuration AttrDict and a debug
    YAML object with comments attached, ready to write to disk.

    Parameters
    ----------
    config : AttrDict
    timeseries_dataframes : dict
    debug_comments : AttrDict
    scenario : str
    """
    model_run = AttrDict()
    model_run["scenario"] = scenario
    model_run["applied_overrides"] = ";".join(applied_overrides)

    # 1) Initial checks on model configuration
    warning_messages, errors = checks.check_initial(config)
    exceptions.print_warnings_and_raise_errors(warnings=warning_messages, errors=errors)

    # 2) Fully populate techs
    # Raises ModelError if necessary
    model_run["techs"], debug_techs, errors = process_techs(config)
    debug_comments.set_key("model_run.techs", debug_techs)
    exceptions.print_warnings_and_raise_errors(errors=errors)

    # 3) Fully populate tech_groups
    model_run["tech_groups"] = process_tech_groups(config, model_run["techs"])

    # 4) Fully populate locations
    (
        model_run["locations"],
        debug_locs,
        warning_messages,
        errors,
    ) = locations.process_locations(config, model_run["techs"])
    debug_comments.set_key("model_run.locations", debug_locs)
    exceptions.print_warnings_and_raise_errors(warnings=warning_messages, errors=errors)

    # 5) Fully populate timeseries data
    # Raises ModelErrors if there are problems with timeseries data at this stage
    model_run["timeseries_data"], model_run["timesteps"] = process_timeseries_data(
        config, model_run, timeseries_dataframes
    )

    # 6) Grab additional relevant bits from run and model config
    model_run["run"] = config["run"]
    model_run["model"] = config["model"]
    model_run["group_constraints"] = config.get("group_constraints", {})

    # 7) Initialize sets
    all_sets = sets.generate_simple_sets(model_run)
    all_sets.union(sets.generate_loc_tech_sets(model_run, all_sets))
    all_sets = AttrDict({k: list(v) for k, v in all_sets.items()})
    model_run["sets"] = all_sets
    model_run["constraint_sets"] = constraint_sets.generate_constraint_sets(model_run)

    # 8) Final sense-checking
    final_check_comments, warning_messages, errors = checks.check_final(model_run)
    debug_comments.union(final_check_comments)
    exceptions.print_warnings_and_raise_errors(warnings=warning_messages, errors=errors)

    # 9) Build a debug data dict with comments and the original configs
    debug_data = AttrDict(
        {
            "comments": debug_comments,
            "config_initial": config,
        }
    )

    return model_run, debug_data



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

clustering.py
~~~~~~~~~~~~~

Functions to cluster data along the time dimension.

"""
import numpy as np
import pandas as pd
import xarray as xr

from sklearn.metrics import mean_squared_error
from sklearn import cluster as sk_cluster

from calliope import exceptions


def _stack_data(data, dates, times):
    """
    Stack all non-time dimensions of an xarray DataArray
    """
    data_to_stack = data.assign_coords(
        timesteps=pd.MultiIndex.from_product([dates, times], names=["dates", "times"])
    ).unstack("timesteps")
    non_date_dims = list(set(data_to_stack.dims).difference(["dates", "times"])) + [
        "times"
    ]
    if len(non_date_dims) >= 2:
        stacked_var = data_to_stack.stack(stacked=non_date_dims)
    else:
        raise exceptions.ModelError(
            "Cannot conduct time clustering with variable {} as it has no "
            "non-time dimensions.".format(data.name)
        )
    return stacked_var


def reshape_for_clustering(data, loc_techs=None, variables=None):
    """
    Create an array of timeseries values, where each day has a row of all
    hourly timeseries data from all relevant variables

    Parameters
    ----------
    data : xarray Dataset
        Dataset with all non-time dependent variables removed
    loc_techs : string or list-like, default = None
        If clustering over a subset of loc_techs, they are listed here
    variables : string or list-like, default = None
        If clustering over a subset of timeseries variables, they are listed here

    Returns
    -------
    X : numpy 2darray
        Array with all rows as days and all other data in columns, with NaNs
        converted to zeros
    """
    # to create an array with days as rows, we need to get timesteps per day
    timesteps = data.timesteps.to_index()
    if timesteps.dtype.kind == "M":  # 'M' = datetime format
        dates = np.unique(timesteps.date)
        times = np.unique(timesteps.time)
    else:  # mean_data from get_closest_days_from_clusters is in the format cluster-timestep
        dates = np.unique([i.split("-")[0] for i in timesteps])
        times = np.unique([i.split("-")[1] for i in timesteps])

    reshaped_data = np.array([[] for i in dates])
    # if 'variables' is given then we will loop over that, otherwise we loop over
    # all timeseries variables
    relevent_vars = variables if variables else data.data_vars

    # loop over variables to get our arrays
    for var in relevent_vars:
        temp_data = data[var].copy()
        # if there is a loc_tech subset, index over that
        if loc_techs:
            loc_tech_dim = [i for i in data[var].dims if "loc_techs" in i][0]
            relevent_loc_techs = list(
                set(temp_data[loc_tech_dim].values).intersection(loc_techs)
            )
            temp_data = temp_data.loc[{loc_tech_dim: relevent_loc_techs}]

        # stack all non-time dimensions to get one row of data for each timestep
        stacked_var = _stack_data(temp_data, dates, times)
        # reshape the array to split days and timesteps within days, now we have
        # one row of data per day
        reshaped_data = np.concatenate([reshaped_data, stacked_var.values], axis=1)
    # put all the columns of data together, keeping rows as days. Also convert
    # all nans to zeros
    X = np.nan_to_num(reshaped_data)

    return X


def get_mean_from_clusters(data, clusters, timesteps_per_day):
    """
    Clusters are days which are considered similar to each other. Here we find
    the mean value (in each tech dimension) for each cluster and return it

    Parameters
    ----------
    data : xarray Dataset
        Dataset with all non-time dependent variables removed
    clusters : pandas DataFrame
        index as days, columns as cluster group to which that day is associated
    timesteps_per_day : int
        Number of timesteps in each day

    Returns
    -------
    ds : xarray Dataset
        Dataset of timeseries DataArrays. Time dimension is only equal to the
        number of clusters * timesteps_per_day.
    """

    cluster_map = clusters.groupby(clusters).groups

    ds = {}
    t_coords = [
        "{}-{}".format(cid, t) for cid in cluster_map for t in range(timesteps_per_day)
    ]
    for var in data.data_vars:
        clustered_array = (
            data[var].loc[{"timesteps": data.timesteps[: len(t_coords)]}].copy()
        )
        clustered_array["timesteps"] = t_coords
        for cluster_id, cluster_members in cluster_map.items():
            current_cluster = [
                i for i in t_coords if i.startswith("{}-".format(cluster_id))
            ]
            clustered_array.loc[{"timesteps": current_cluster}] = (
                data[var]
                .loc[{"timesteps": cluster_members}]
                .groupby("timesteps.time")
                .mean(dim="timesteps")
                .values
            )
        ds[var] = clustered_array
    ds = xr.Dataset(ds)
    return ds


def find_nearest_vector_index(array, value, metric="rmse"):
    """
    compares the data for one cluster to every day in the timeseries, to find the
    day which most closely matches.

    Parameters
    ----------
    array : np.ndarray
        full timeseries array,
        shape = (num_dates, num_loc_techs * num_vars * num_timesteps_per_day)
        The shape is acheived by running a dataset through clustering.reshape_for_clustering
    value : np.ndarray
        one cluster of data,
        shape = (1, num_loc_techs * num_vars * num_timesteps_per_day)
        The shape is acheived by running the mean clustered dataset,
        subset over one cluster, through clustering.reshape_for_clustering
    metric : str, default = 'rmse'
        Error metric with which to compare the array and the values.
        If 'rmse', will compare the root mean square error between `value` and
        each date in `array`.
        If 'abs' will compare the absolute difference between `value` and
        each date in `array`.
    """
    if metric == "rmse":
        error = np.array([mean_squared_error(i, value[0]) for i in array])
    elif metric == "abs":
        error = np.array([np.linalg.norm(sum(y)) for y in array - value])
    else:
        raise ValueError(
            "Error metric can only be `rmse` or `abs`, {} given".format(metric)
        )

    return error.argmin()


def get_closest_days_from_clusters(data, mean_data, clusters, daily_timesteps):
    """
    Given a set of mean cluster timeseries profiles, find the day in the full
    timeseries that matches each cluster profile most closely.

    Parameters
    ----------
    data : xarray Dataset
        Input model data, with all non-timeseries data vars dropped
    mean_data : xarray Dataset
        Mean per cluster of input model data, resulting timeseries is of the form
        X-Y where X is the cluster number and Y is the timestep (both integers)
    clusters : pd.Series
        Cluster to which each date is aligned.
        Index = dates of the full timeseries, values = cluster number
    daily_timesteps : list
        hours assigned to each timestep in a day (e.g. an entry of 0.25 = 15 minutes).
        We expect uniform timesteps between days, which is checked prior to
        reaching this function.

    Returns
        new_data : xarray Dataset
            data, indexed over only the timesteps within the days that most
            closely match the cluster means.
            Number of timesteps = timesteps_per_day * num_clusters. May be lower
            than this if many cluster means match with the same day
        chosen_day_timestepas : dict
            The day assigned to each cluster. key = cluster number, value = date.
    """
    dtindex = data["timesteps"].to_index()
    timesteps_per_day = len(daily_timesteps)
    chosen_days = {}

    for cluster in sorted(clusters.unique()):

        subset_t = [
            t for t in mean_data.timesteps.values if t.startswith("{}-".format(cluster))
        ]

        target = reshape_for_clustering(mean_data.loc[dict(timesteps=subset_t)])

        lookup_array = reshape_for_clustering(data)

        chosen_days[cluster] = find_nearest_vector_index(lookup_array, target)

    days_list = sorted(list(set(chosen_days.values())))
    new_t_coord = _timesteps_from_daily_index(
        dtindex[::timesteps_per_day][days_list], daily_timesteps
    )

    chosen_day_timestamps = {
        k: dtindex[::timesteps_per_day][v] for k, v in chosen_days.items()
    }

    new_data = data.loc[dict(timesteps=new_t_coord)]

    return new_data, chosen_day_timestamps


def _timesteps_from_daily_index(idx, daily_timesteps):
    new_idx = pd.Index(
        [
            date + pd.DateOffset(hours=sum(daily_timesteps[:timestep]))
            for date in idx
            for timestep in range(len(daily_timesteps))
        ]
    )
    return new_idx


def map_clusters_to_data(
    data, clusters, how, daily_timesteps, storage_inter_cluster=True
):
    """
    Returns a copy of data that has been clustered.

    Parameters
    ----------
    how : str
        How to select data from clusters. Can be mean (centroid) or closest real
        day to the mean (by root mean square error).
    storage_inter_cluster : bool, default=True
        If True, add `datesteps` to model_data, for use in the backend to build
        inter_cluster storage decision variables and constraints
    """
    # FIXME hardcoded time intervals ('1H', '1D')

    # Get all timesteps, not just the first per day
    timesteps_per_day = len(daily_timesteps)
    idx = clusters.index
    new_idx = _timesteps_from_daily_index(idx, daily_timesteps)
    clusters_timeseries = clusters.reindex(new_idx).fillna(method="ffill").astype(int)

    new_data = get_mean_from_clusters(data, clusters_timeseries, timesteps_per_day)
    new_data.attrs = data.attrs

    if how == "mean":
        # Add timestep names by taking the median timestamp from daily clusters...
        # (a random way of doing it, but we want some label to apply)
        timestamps = clusters.groupby(clusters).apply(
            lambda x: x.index[int(len(x.index) / 2)]
        )
        new_data.coords["timesteps"] = _timesteps_from_daily_index(
            pd.Index(timestamps.values), daily_timesteps
        )
        # Generate weights
        # weight of each timestep = number of timesteps in this timestep's cluster
        # divided by timesteps per day (since we're grouping days together and
        # a cluster consisting of 1 day = 24 hours should have weight of 1)
        value_counts = clusters_timeseries.value_counts() / timesteps_per_day
        # And turn the index into dates (days)
        value_counts = pd.DataFrame(
            {"dates": timestamps, "counts": value_counts}
        ).set_index("dates")["counts"]

    elif how == "closest":
        new_data, chosen_ts = get_closest_days_from_clusters(
            data, new_data, clusters, daily_timesteps
        )
        # Deal with the case where more than one cluster has the same closest day
        # An easy way is to rename the original clusters with the chosen days
        # So at this point, clusterdays_timeseries maps all timesteps to the day
        # of year of the cluster the timestep belongs to
        clusterdays_timeseries = clusters_timeseries.map(lambda x: chosen_ts[x])
        value_counts = clusterdays_timeseries.value_counts() / timesteps_per_day
        timestamps = pd.DataFrame.from_dict(chosen_ts, orient="index")[0]
        cluster_diff = len(clusters.unique()) - len(timestamps.unique())
        if cluster_diff > 0:
            exceptions.warn(
                "Creating {} fewer clusters as some clusters share the same "
                "closest day".format(cluster_diff)
            )
            timestamps = timestamps.drop_duplicates()
            for cluster, date in timestamps.items():
                clusterdays_timeseries.loc[clusterdays_timeseries == date] = cluster
            clusters = clusterdays_timeseries.astype(int).resample("1D").mean()

    _clusters = xr.DataArray(
        data=np.full(len(new_data.timesteps.values), np.nan),
        dims="timesteps",
        coords={"timesteps": new_data.timesteps.values},
    )

    for cluster, date in timestamps.items():
        _clusters.loc[date.strftime("%Y-%m-%d")] = cluster

    new_data["timestep_cluster"] = _clusters.astype(int)
    weights = value_counts.reindex(
        _timesteps_from_daily_index(value_counts.index, daily_timesteps)
    ).fillna(method="ffill")
    new_data["timestep_weights"] = xr.DataArray(weights, dims=["timesteps"])
    days = np.unique(new_data.timesteps.to_index().date)
    new_data["timestep_resolution"] = xr.DataArray(
        np.tile(daily_timesteps, len(days)),
        dims=["timesteps"],
        coords={"timesteps": new_data["timesteps"]},
    )

    if storage_inter_cluster:
        clusters.index.name = "datesteps"
        new_data["lookup_datestep_cluster"] = xr.DataArray.from_series(clusters)
    timestamps.index.name = "clusters"
    new_data.coords["clusters"] = timestamps.index

    return new_data


def get_clusters(
    data,
    func,
    timesteps_per_day,
    tech=None,
    timesteps=None,
    k=None,
    variables=None,
    **kwargs,
):
    """
    Run a clustering algorithm on the timeseries data supplied. All timeseries
    data is reshaped into one row per day before clustering into similar days.

    Parameters
    ----------
    data : xarray.Dataset
        Should be normalized
    func : str
        'kmeans' or 'hierarchical' for KMeans or Agglomerative clustering, respectively
    timesteps_per_day : int
        Total number of timesteps in a day
    tech : list, optional
        list of strings referring to technologies by which clustering is undertaken.
        If none (default), all technologies within timeseries variables will be used.
    timesteps : list or str, optional
        Subset of the time domain within which to apply clustering.
    k : int, optional
        Number of clusters to create. If none (default), will use Hartigan's rule
        to infer a reasonable number of clusters.
    variables : list, optional
        data variables (e.g. `resource`, `energy_eff`) by whose values the data
        will be clustered. If none (default), all timeseries variables will be used.
    kwargs : dict
        Additional keyword arguments available depend on the `func`.
        For available KMeans kwargs see:
        http://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html
        For available hierarchical kwargs see:
        http://scikit-learn.org/stable/modules/generated/sklearn.cluster.AgglomerativeClustering.html
    Returns
    -------
    clusters : dataframe
        Indexed by timesteps and with locations as columns, giving cluster
        membership for first timestep of each day.
    clustered_data : sklearn.cluster object
        Result of clustering using sklearn.KMeans(k).fit(X) or
        sklearn.KMeans(k).AgglomerativeClustering(X). Allows user to access
        specific attributes, for detailed statistical analysis.

    """

    if timesteps is not None:
        data = data.loc[{"timesteps": timesteps}]
    else:
        timesteps = data.timesteps.values

    X = reshape_for_clustering(data, tech, variables)

    if func == "kmeans":
        if not k:
            k = hartigan_n_clusters(X)
            exceptions.warn(
                "Used Hartigan's rule to determine that"
                "a good number of clusters is {}.".format(k)
            )
        clustered_data = sk_cluster.KMeans(k).fit(X)

    elif func == "hierarchical":
        if not k:
            raise exceptions.ModelError(
                "Cannot undertake hierarchical clustering without a predefined "
                "number of clusters (k)"
            )
        clustered_data = sk_cluster.AgglomerativeClustering(k).fit(X)

    # Determine the cluster membership of each day
    day_clusters = clustered_data.labels_

    # Create mapping of timesteps to clusters
    clusters = pd.Series(day_clusters, index=timesteps[::timesteps_per_day])

    return clusters, clustered_data


def hartigan_n_clusters(X, threshold=10):
    """
    Try clustering using sklearn.cluster.kmeans, for several cluster sizes.
    Using Hartigan's rule, we will return the number of clusters after which
    the benefit of clustering is low.
    """

    def _H_rule(inertia, inertia_plus_one, n_clusters, len_input):
        # see http://www.dcs.bbk.ac.uk/~mirkin/papers/00357_07-216RR_mirkin.pdf
        return ((inertia / inertia_plus_one) - 1) * (len_input - n_clusters - 1)

    len_input = len(X)
    n_clusters = 1
    HK = threshold + 1

    while n_clusters <= len_input and HK > threshold:

        kmeans = sk_cluster.KMeans(n_clusters=n_clusters).fit(X)
        kmeans_plus_one = sk_cluster.KMeans(n_clusters=n_clusters + 1).fit(X)

        inertia = kmeans.inertia_
        inertia_plus_one = kmeans_plus_one.inertia_

        HK = _H_rule(inertia, inertia_plus_one, n_clusters, len_input)

        n_clusters += 1

    if HK > threshold:  # i.e. we went to the limit where n_clusters = len_input
        exceptions.warn("Based on threshold, number of clusters = number of dates")
        return len_input
    else:
        return n_clusters - 1



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

import logging
import re

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
import xarray as xr
import pyomo.core as po

from calliope.core.util.tools import memoize
from calliope import exceptions

logger = logging.getLogger(__name__)


@memoize
def get_param(backend_model, var, dims):
    """
    Get an input parameter held in a Pyomo object, or held in the defaults
    dictionary if that Pyomo object doesn't exist.

    Parameters
    ----------
    backend_model : Pyomo model instance
    var : str
    dims : single value or tuple

    """
    try:
        return getattr(backend_model, var)[dims]
    except AttributeError:  # i.e. parameter doesn't exist at all
        logger.debug(
            "get_param: var {} and dims {} leading to default lookup".format(var, dims)
        )
        return backend_model.__calliope_defaults[var]
    except KeyError:  # try removing timestep
        try:
            if len(dims) > 2:
                return getattr(backend_model, var)[dims[:-1]]
            else:
                return getattr(backend_model, var)[dims[0]]
        except KeyError:  # Static default value
            logger.debug(
                "get_param: var {} and dims {} leading to default lookup".format(
                    var, dims
                )
            )
            return backend_model.__calliope_defaults[var]


def get_previous_timestep(timesteps, timestep):
    """Get the timestamp for the timestep previous to the input timestep"""
    return timesteps[timesteps.ord(timestep) - 1]


@memoize
def get_loc_tech_carriers(backend_model, loc_carrier):
    """
    For a given loc_carrier concatenation, get lists of the relevant
    loc_tech_carriers which produce energy (loc_tech_carriers_prod), consume
    energy (loc_tech_carriers_con) and export energy (loc_tech_carriers_export)
    """
    lookup = backend_model.__calliope_model_data["data"]["lookup_loc_carriers"]
    loc_tech_carriers = split_comma_list(lookup[loc_carrier])

    loc_tech_carriers_prod = [
        i for i in loc_tech_carriers if i in backend_model.loc_tech_carriers_prod
    ]
    loc_tech_carriers_con = [
        i for i in loc_tech_carriers if i in backend_model.loc_tech_carriers_con
    ]

    if hasattr(backend_model, "loc_tech_carriers_export"):
        loc_tech_carriers_export = [
            i for i in loc_tech_carriers if i in backend_model.loc_tech_carriers_export
        ]
    else:
        loc_tech_carriers_export = []

    return (loc_tech_carriers_prod, loc_tech_carriers_con, loc_tech_carriers_export)


@memoize
def get_loc_tech(loc_tech_carrier):
    """
    Split the string of a loc_tech_carrier (e.g. `region1::ccgt::power`) to get
    just the loc_tech (e.g. `region1::ccgt`)
    """
    return loc_tech_carrier.rsplit("::", 1)[0]


@memoize
def get_timestep_weight(backend_model):
    """
    Get the total number of years this model considers, by summing all
    timestep resolution with timestep weight (a weight/resolution of 1 = 1 hour)
    and divide it by number of hours in the year. Weight/resolution will almost
    always be 1 per step, unless time clustering/masking/resampling has taken place.
    """
    model_data_dict = backend_model.__calliope_model_data
    time_res = list(model_data_dict["data"]["timestep_resolution"].values())
    weights = list(model_data_dict["data"]["timestep_weights"].values())
    return sum(np.multiply(time_res, weights)) / 8760


@memoize
def split_comma_list(comma_list):
    """
    Take a comma deliminated string and split it into a list of strings
    """
    return comma_list.split(",")


@memoize
def get_conversion_plus_io(backend_model, tier):
    """
    from a carrier_tier, return the primary tier (of `in`, `out`) and
    corresponding decision variable (`carrier_con` and `carrier_prod`, respectively)
    """
    if "out" in tier:
        return "out", backend_model.carrier_prod
    elif "in" in tier:
        return "in", backend_model.carrier_con


def get_var(backend_model, var, dims=None, sparse=False, expr=False):
    """
    Return output for variable `var` as a pandas.Series (1d),
    pandas.Dataframe (2d), or xarray.DataArray (3d and higher).

    Parameters
    ----------
    var : variable name as string, e.g. 'resource'
    dims : list, optional
        indices as strings, e.g. ('loc_techs', 'timesteps');
        if not given, they are auto-detected
    sparse : bool, optional; default = False
        If extracting Pyomo Param data, the output sparse array includes inputs
        the user left as NaN replaced with the default value for that Param.
    expr : bool, optional
        If True, treat var as a pyomo expression, which requires calculating
        the result of the expression before translating into nd data structure
    """
    try:
        var_container = getattr(backend_model, var)
    except AttributeError:
        raise exceptions.BackendError("Variable {} inexistent.".format(var))

    if not dims:
        if var + "_index" == var_container.index_set().name:
            dims = [i.name for i in var_container.index_set().subsets()]
        else:
            dims = [var_container.index_set().name]

    if sparse and not expr:
        if invalid(var_container.default()):
            result = pd.Series(var_container._data).apply(
                lambda x: po.value(x) if not invalid(x) else np.nan
            )
        else:
            result = pd.Series(var_container.extract_values_sparse())
    else:
        if expr:
            result = pd.Series(var_container._data).apply(po.value)
        else:
            result = pd.Series(var_container.extract_values())
    if result.empty:
        raise exceptions.BackendError("Variable {} has no data.".format(var))

    result = result.rename_axis(index=dims)

    return xr.DataArray.from_series(result)


@memoize
def loc_tech_is_in(backend_model, loc_tech, model_set):
    """
    Check if set exists and if loc_tech is in the set

    Parameters
    ----------
    loc_tech : string
    model_set : string
    """

    if hasattr(backend_model, model_set) and loc_tech in getattr(
        backend_model, model_set
    ):
        return True
    else:
        return False


def get_domain(var: xr.DataArray) -> str:
    def check_sign(var):
        if re.match("resource|loc_coordinates|cost*", var.name):
            return ""
        elif re.match("group_carrier_con*", var.name):
            return "NonPositive"
        else:
            return "NonNegative"

    if var.dtype.kind == "b":
        return "Boolean"
    elif is_numeric_dtype(var.dtype):
        return check_sign(var) + "Reals"
    else:
        return "Any"


def invalid(val) -> bool:
    if isinstance(val, po.base.param._ParamData):
        return val._value == po.Param.NoValue or po.value(val) is None
    elif val == po.Param.NoValue:
        return True
    else:
        return pd.isnull(val)


def datetime_to_string(
    backend_model: po.ConcreteModel, model_data: xr.Dataset
) -> xr.Dataset:
    """
    Convert from datetime to string xarray dataarrays, to reduce the memory
    footprint of converting datetimes from numpy.datetime64 -> pandas.Timestamp
    when creating the pyomo model object.
    Parameters
    ----------
    backend_model : the backend pyomo model object
    model_data : the Calliope xarray Dataset of model data
    """
    datetime_data = set()
    for attr in ["coords", "data_vars"]:
        for set_name, set_data in getattr(model_data, attr).items():
            if set_data.dtype.kind == "M":
                attrs = model_data[set_name].attrs
                model_data[set_name] = model_data[set_name].dt.strftime(
                    "%Y-%m-%d %H:%M"
                )
                model_data[set_name].attrs = attrs
                datetime_data.add((attr, set_name))
    backend_model.__calliope_datetime_data = datetime_data

    return model_data


def string_to_datetime(
    backend_model: po.ConcreteModel, model_data: xr.Dataset
) -> xr.Dataset:
    """
    Convert from string to datetime xarray dataarrays, reverting the process
    undertaken in
    datetime_to_string
    Parameters
    ----------
    backend_model : the backend pyomo model object
    model_data : the Calliope xarray Dataset of model data
    """
    for attr, set_name in backend_model.__calliope_datetime_data:
        if attr == "coords" and set_name in model_data:
            model_data.coords[set_name] = model_data[set_name].astype("datetime64[ns]")
        elif set_name in model_data:
            model_data[set_name] = (
                model_data[set_name].fillna(pd.NaT).astype("datetime64[ns]")
            )
    return model_data



import pytest  # pylint: disable=unused-import

from calliope.test.common.util import build_test_model as build_model
from calliope.test.common.util import check_variable_exists


class TestBuildConversionPlusConstraints:
    # conversion_plus.py
    def test_loc_techs_balance_conversion_plus_primary_constraint(self):
        """
        sets.loc_techs_conversion_plus,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "balance_conversion_plus_primary_constraint"
        )

        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_primary_constraint")

        m = build_model(
            {
                "techs.test_conversion_plus.essentials.carrier_out": [
                    "electricity",
                    "heat",
                ]
            },
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_primary_constraint")

        m = build_model(
            {
                "techs.test_conversion_plus.essentials": {
                    "carrier_in": ["coal", "gas"],
                    "primary_carrier_in": "gas",
                }
            },
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_primary_constraint")

    def test_loc_techs_carrier_production_max_conversion_plus_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if i not in sets.loc_techs_milp
        """

        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "carrier_production_max_conversion_plus_constraint"
        )

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_techs_carrier_production_max_conversion_plus_milp_constraint(self):
        m = build_model({}, "conversion_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_max_conversion_plus_constraint"
        )

    def test_loc_techs_carrier_production_min_conversion_plus_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i not in sets.loc_techs_milp
        """

        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_constraint"
        )

        m = build_model(
            {"techs.test_conversion_plus.constraints.energy_cap_min_use": 0.1},
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_constraint"
        )

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_techs_carrier_production_min_conversion_plus_milp_constraint(self):
        m = build_model({}, "conversion_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_constraint"
        )

        m = build_model(
            {"techs.test_conversion_plus.constraints.energy_cap_min_use": 0.1},
            "conversion_plus_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_constraint"
        )

    def test_loc_techs_cost_var_conversion_plus_constraint(self):
        """
        sets.loc_techs_om_cost_conversion_plus,
        """
        # no conversion_plus = no constraint
        m = build_model(
            {"techs.test_supply_elec.costs.monetary.om_prod": 0.1},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "cost_var_conversion_plus_constraint")

        # no conversion_plus = no constraint
        m = build_model(
            {"techs.test_conversion.costs.monetary.om_prod": 0.1},
            "simple_conversion,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "cost_var_conversion_plus_constraint")

        # no variable costs for conversion_plus = no constraint
        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "cost_var_conversion_plus_constraint")

        # om_prod creates constraint and populates it with carrier_prod driven cost
        m = build_model(
            {"techs.test_conversion_plus.costs.monetary.om_prod": 0.1},
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_var_conversion_plus_constraint")
        assert check_variable_exists(
            m._backend_model, "cost_var_conversion_plus_constraint", "carrier_prod"
        )
        assert not check_variable_exists(
            m._backend_model, "cost_var_conversion_plus_constraint", "carrier_con"
        )

        # om_con creates constraint and populates it with carrier_con driven cost
        m = build_model(
            {"techs.test_conversion_plus.costs.monetary.om_con": 0.1},
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_var_conversion_plus_constraint")
        assert check_variable_exists(
            m._backend_model, "cost_var_conversion_plus_constraint", "carrier_con"
        )
        assert not check_variable_exists(
            m._backend_model, "cost_var_conversion_plus_constraint", "carrier_prod"
        )

    def test_loc_techs_balance_conversion_plus_in_2_constraint(self):
        """
        sets.loc_techs_in_2,
        """

        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "balance_conversion_plus_in_2_constraint")

        m = build_model(
            {
                "techs.test_conversion_plus.essentials": {
                    "carrier_in_2": "coal",
                    "primary_carrier_in": "gas",
                }
            },
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_in_2_constraint")

        m = build_model(
            {
                "techs.test_conversion_plus.essentials": {
                    "carrier_in_2": ["coal", "heat"],
                    "primary_carrier_in": "gas",
                }
            },
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_in_2_constraint")

    def test_loc_techs_balance_conversion_plus_in_3_constraint(self):
        """
        sets.loc_techs_in_3,
        """

        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "balance_conversion_plus_in_3_constraint")

        m = build_model(
            {
                "techs.test_conversion_plus.essentials": {
                    "carrier_in_3": "coal",
                    "primary_carrier_in": "gas",
                }
            },
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_in_3_constraint")

        m = build_model(
            {
                "techs.test_conversion_plus.essentials": {
                    "carrier_in_3": ["coal", "heat"],
                    "primary_carrier_in": "gas",
                }
            },
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_in_3_constraint")

    def test_loc_techs_balance_conversion_plus_out_2_constraint(self):
        """
        sets.loc_techs_out_2,
        """

        m = build_model(
            {"techs.test_conversion_plus.essentials.carrier_out_2": ["coal", "heat"]},
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_out_2_constraint")

    def test_loc_techs_balance_conversion_plus_out_3_constraint(self):
        """
        sets.loc_techs_out_3,
        """

        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "balance_conversion_plus_out_3_constraint")

        m = build_model(
            {"techs.test_conversion_plus.essentials.carrier_out_3": "coal"},
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_out_3_constraint")

        m = build_model(
            {"techs.test_conversion_plus.essentials.carrier_out_3": ["coal", "heat"]},
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_plus_out_3_constraint")

    def test_loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint(self):
        """ """

        m = build_model({}, "simple_conversion_plus,one_day,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "conversion_plus_prod_con_to_zero_constraint"
        )

        m = build_model(
            {
                "techs.test_conversion_plus.constraints.carrier_ratios.carrier_out_2.heat": "file=carrier_ratio.csv"
            },
            "simple_conversion_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "conversion_plus_prod_con_to_zero_constraint"
        )

        m = build_model(
            {
                "techs.test_conversion_plus.constraints.carrier_ratios.carrier_out_2.heat": "file=carrier_ratio.csv"
            },
            "simple_conversion_plus,one_day,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "conversion_plus_prod_con_to_zero_constraint")
        assert check_variable_exists(
            m._backend_model,
            "conversion_plus_prod_con_to_zero_constraint",
            "carrier_prod",
        )
        assert hasattr(
            m._backend_model,
            "loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint",
        )

        m = build_model(
            {
                "techs.test_conversion_plus": {
                    "essentials.carrier_in": ["gas", "coal"],
                    "essentials.primary_carrier_in": "gas",
                    "constraints.carrier_ratios": {
                        "carrier_out_2.heat": "file=carrier_ratio.csv",
                        "carrier_in.coal": "file=carrier_ratio.csv",
                    },
                }
            },
            "simple_conversion_plus,one_day,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "conversion_plus_prod_con_to_zero_constraint")
        assert check_variable_exists(
            m._backend_model,
            "conversion_plus_prod_con_to_zero_constraint",
            "carrier_prod",
        )
        assert check_variable_exists(
            m._backend_model,
            "conversion_plus_prod_con_to_zero_constraint",
            "carrier_con",
        )
        assert hasattr(
            m._backend_model,
            "loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint",
        )


class TestConversionPlusConstraintResults:
    def test_carrier_ratio_from_file(self):
        m = build_model(
            {"techs.test_conversion.costs.monetary.om_prod": 3},
            "conversion_and_conversion_plus,one_day,investment_costs",
        )
        m.run()
        carrier_prod_conversion_plus = (
            (
                m.get_formatted_array("carrier_prod").loc[
                    {"techs": "test_conversion_plus"}
                ]
            )
            .sum("locs")
            .to_pandas()
        )
        assert all(
            (
                carrier_prod_conversion_plus.loc["heat"]
                / carrier_prod_conversion_plus.loc["electricity"]
            ).dropna()
            == 0.8
        )

        m = build_model(
            {
                "techs.test_conversion_plus": {
                    "constraints.carrier_ratios": {
                        "carrier_out_2.heat": "file=carrier_ratio.csv",
                        "carrier_in.coal": "file=carrier_ratio.csv",
                    },
                    "essentials": {
                        "carrier_in": ["gas", "coal"],
                        "primary_carrier_in": "gas",
                    },
                    "costs.monetary": {"om_con": 3, "om_prod": 3},
                }
            },
            "conversion_and_conversion_plus,one_day,investment_costs",
        )
        m.run()
        carrier_prod_conversion_plus = (
            (
                m.get_formatted_array("carrier_prod").loc[
                    {"techs": "test_conversion_plus"}
                ]
            )
            .sum("locs")
            .to_pandas()
        )
        carrier_con_conversion_plus = (
            (
                m.get_formatted_array("carrier_con").loc[
                    {"techs": "test_conversion_plus"}
                ]
            )
            .sum("locs")
            .to_pandas()
        )

        prod_ratios = (
            carrier_prod_conversion_plus.loc["heat"]
            / carrier_prod_conversion_plus.loc["electricity"]
        ).dropna()
        con_ratios = (
            carrier_con_conversion_plus.loc["coal"]
            / carrier_prod_conversion_plus.sum(axis=0)
        ).dropna()

        assert (
            m._model_run.timeseries_data["carrier_ratio.csv"]
            .loc[:, "0"]
            .reindex(prod_ratios.index)
            == prod_ratios.round(1)
        ).all()
        assert (
            m._model_run.timeseries_data["carrier_ratio.csv"]
            .loc[:, "0"]
            .reindex(con_ratios.index)
            == con_ratios.round(1)
        ).all()



import pytest
from pytest import approx

import pyomo.core as po

import calliope
from calliope.test.common.util import build_test_model as build_model


class TestCostMinimisationObjective:
    def test_nationalscale_minimize_emissions(self):
        model = calliope.examples.national_scale(
            scenario="minimize_emissions_costs",
            override_dict={"model.subset_time": "2005-01-01"},
        )
        model.run()

        assert model.results.energy_cap.to_pandas()["region1-2::csp"] == approx(10000.0)
        assert model.results.energy_cap.to_pandas()[
            "region1::ac_transmission:region2"
        ] == approx(10000.0)

        assert model.results.carrier_prod.sum("timesteps").to_pandas()[
            "region1::ccgt::power"
        ] == approx(66530.36492823533)

        assert float(model.results.cost.loc[{"costs": "emissions"}].sum()) == approx(
            13129619.1
        )

    def test_nationalscale_maximize_utility(self):
        model = calliope.examples.national_scale(
            scenario="maximize_utility_costs",
            override_dict={"model.subset_time": "2005-01-01"},
        )
        model.run()

        assert model.results.energy_cap.to_pandas()["region1-2::csp"] == approx(10000.0)
        assert model.results.energy_cap.to_pandas()[
            "region1::ac_transmission:region2"
        ] == approx(10000.0)

        assert model.results.carrier_prod.sum("timesteps").to_pandas()[
            "region1::ccgt::power"
        ] == approx(115569.4354)

        assert float(model.results.cost.sum()) > 6.6e7

    @pytest.mark.parametrize(
        "scenario,cost_class,weight",
        [
            ("monetary_objective", ["monetary"], [1]),
            ("emissions_objective", ["emissions"], [1]),
            ("weighted_objective", ["monetary", "emissions"], [0.9, 0.1]),
        ],
    )
    def test_weighted_objective_results(self, scenario, cost_class, weight):
        model = build_model(model_file="weighted_obj_func.yaml", scenario=scenario)
        model.run()
        assert sum(
            model.results.cost.loc[{"costs": cost_class[i]}].sum().item() * weight[i]
            for i in range(len(cost_class))
        ) == approx(po.value(model._backend_model.obj))

    def test_update_cost_classes_weights(self):
        model = build_model(
            model_file="weighted_obj_func.yaml", scenario="weighted_objective"
        )

        model.run()
        obj_value = model._backend_model.obj()
        total_cost = model.results.cost.sum()
        model.backend.update_param("objective_cost_class", {"monetary": 1.8})
        model.backend.update_param("objective_cost_class", {"emissions": 0.2})

        new_model = model.backend.rerun()
        updated_obj_value = model._backend_model.obj()
        updated_total_cost = new_model.results.cost.sum()

        assert updated_obj_value == 2 * obj_value
        assert updated_total_cost == total_cost



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""
import logging

import numpy as np
import pandas as pd
import xarray as xr
from calliope import exceptions
from calliope.backend import checks
from calliope.backend.pyomo import interface as pyomo_interface
from calliope.backend.pyomo import model as run_pyomo
from calliope.core import io
from calliope.core.attrdict import AttrDict
from calliope.core.util.dataset import split_loc_techs
from calliope.core.util.logging import log_time
from calliope.core.util.observed_dict import UpdateObserverDict

logger = logging.getLogger(__name__)


def run(model_data, timings, build_only=False):
    """
    Parameters
    ----------
    model_data : xarray.Dataset
        Pre-processed dataset of Calliope model data.
    timings : dict
        Stores timings of various stages of model processing.
    build_only : bool, optional
        If True, the backend only constructs its in-memory representation
        of the problem rather than solving it. Used for debugging and
        testing.

    """

    BACKEND = {"pyomo": run_pyomo}

    INTERFACE = {"pyomo": pyomo_interface}

    run_config = AttrDict.from_yaml_string(model_data.attrs["run_config"])

    if run_config["mode"] == "plan":
        results, backend, opt = run_plan(
            model_data,
            timings,
            backend=BACKEND[run_config.backend],
            build_only=build_only,
        )

    elif run_config["mode"] == "operate":
        results, backend, opt = run_operate(
            model_data,
            timings,
            backend=BACKEND[run_config.backend],
            build_only=build_only,
        )

    elif run_config["mode"] == "spores":
        results, backend, opt = run_spores(
            model_data,
            timings,
            interface=INTERFACE[run_config.backend],
            backend=BACKEND[run_config.backend],
            build_only=build_only,
        )

    return results, backend, opt, INTERFACE[run_config.backend].BackendInterfaceMethods


def run_plan(
    model_data,
    timings,
    backend,
    build_only,
    backend_rerun=False,
    allow_warmstart=False,
    persistent=True,
    opt=None,
):

    log_time(logger, timings, "run_start", comment="Backend: starting model run")

    warmstart = False
    if not backend_rerun:
        backend_model = backend.generate_model(model_data)
        log_time(
            logger,
            timings,
            "run_backend_model_generated",
            time_since_run_start=True,
            comment="Backend: model generated",
        )

    else:
        backend_model = backend_rerun
        if allow_warmstart:
            warmstart = True

    run_config = backend_model.__calliope_run_config
    solver = run_config["solver"]
    solver_io = run_config.get("solver_io", None)
    solver_options = run_config.get("solver_options", None)
    save_logs = run_config.get("save_logs", None)

    if build_only:
        results = xr.Dataset()

    else:
        if "persistent" in solver and persistent is False:
            exceptions.warn(
                f"The chosen solver, `{solver}` will not be used in this run. "
                f"`{solver.replace('_persistent', '')}` will be used instead."
            )
            solver = solver.replace("_persistent", "")
        log_time(
            logger,
            timings,
            "run_solver_start",
            comment="Backend: sending model to solver",
        )

        backend_results, opt = backend.solve_model(
            backend_model,
            solver=solver,
            solver_io=solver_io,
            solver_options=solver_options,
            save_logs=save_logs,
            warmstart=warmstart,
            opt=opt,
        )

        log_time(
            logger,
            timings,
            "run_solver_exit",
            time_since_run_start=True,
            comment="Backend: solver finished running",
        )

        termination = backend.load_results(backend_model, backend_results, opt)

        log_time(
            logger, timings, "run_results_loaded", comment="Backend: loaded results"
        )

        if termination in ["optimal", "feasible"]:
            results = backend.get_result_array(backend_model, model_data)
            results.attrs["termination_condition"] = termination
            if "persistent" in opt.name and persistent is True:
                results.attrs["objective_function_value"] = opt.get_model_attr("ObjVal")
            else:
                results.attrs["objective_function_value"] = backend_model.obj()
        else:
            results = xr.Dataset(attrs={"termination_condition": termination})

        log_time(
            logger,
            timings,
            "run_solution_returned",
            time_since_run_start=True,
            comment="Backend: generated solution array",
        )

    return results, backend_model, opt


def run_spores(
    model_data, timings, interface, backend, build_only, backend_rerun=False, opt=None
):
    """
    For use when mode is 'spores', to allow the model to be built, edited, and
    iteratively run within Pyomo, modifying, at each iteration, the score of
    each loc::tech in such a way to generate Spatially explicit Practically Optimal
    RESults (SPORES).

    """
    log_time(
        logger,
        timings,
        "run_start",
        comment="Backend: starting model run in SPORES mode",
    )

    def _cap_loc_score_default(results):
        # Define default scoring function, based on integer scoring method
        # TODO: make it possible to point to a custom function instead of this one
        cap_loc_score = xr.where(results.energy_cap > 1e-3, 100, 0).loc[
            {"loc_techs": results.loc_techs_investment_cost}
        ]
        return cap_loc_score.to_series().rename_axis(index="loc_techs_investment_cost")

    # Define function to update "spores_score" after each iteration of the method
    def _update_spores_score(backend_model, cap_loc_score):
        print("Updating loc::tech spores scores")
        loc_tech_score_dict = cap_loc_score.to_dict()

        interface.update_pyomo_param(
            backend_model, opt, "cost_energy_cap", loc_tech_score_dict
        )

    def _warn_on_infeasibility():
        return exceptions.warn(
            "Infeasible SPORE detected. Please check your model configuration. "
            "No more SPORES will be generated."
        )

    def _get_updated_spores_inputs(backend_model):
        inputs = interface.access_pyomo_model_inputs(backend_model)
        inputs_to_keep = ["cost_energy_cap", "group_cost_max"]
        return inputs[inputs_to_keep]

    def _combine_spores_results_and_inputs(
        backend_model, results, spore_num, model_data=None
    ):
        inputs_to_keep = _get_updated_spores_inputs(backend_model)
        for var_data in results.data_vars.values():
            if "is_result" not in var_data.attrs.keys():
                var_data.attrs["is_result"] = 1
        if model_data is not None:
            datasets = [inputs_to_keep, model_data, results]
        else:
            datasets = [inputs_to_keep, results]
        new_ds = xr.combine_by_coords(
            datasets, compat="override", combine_attrs="no_conflicts"
        )
        return new_ds.assign_coords(spores=("spores", [spore_num]))

    def _add_results_to_list(backend_model, spores_results, results, spore_num):
        results_to_add = _combine_spores_results_and_inputs(
            backend_model, results, spore_num
        )
        spores_results[spore_num] = results_to_add

    def _save_spore(backend_model, results, spore_num, model_data=None):
        _path = spores_config["save_per_spore_path"].format(spore_num)
        new_ds = _combine_spores_results_and_inputs(
            backend_model, results, spore_num, model_data
        )
        print(f"Saving SPORE {spore_num} to {_path}")
        io.save_netcdf(new_ds, _path)

    def _update_slack_cost_constraint(backend_model):
        slack_costs = model_data.group_cost_max.loc[
            {"group_names_cost_max": spores_config["slack_cost_group"]}
        ].dropna("costs")
        interface.update_pyomo_param(
            backend_model,
            opt,
            "group_cost_max",
            {
                (_cost_class, spores_config["slack_cost_group"]): results.cost.loc[
                    {"costs": _cost_class}
                ]
                .sum()
                .item()
                * (1 + spores_config["slack"])
                for _cost_class in slack_costs.costs.values
            },
        )

    def _update_to_spores_objective(backend_model):
        interface.update_pyomo_param(
            backend_model,
            opt,
            "objective_cost_class",
            spores_config["objective_cost_class"],
        )

    def _initialise_backend_model():
        if backend_rerun:
            kwargs = {"backend_rerun": backend_rerun, "opt": opt}
        else:
            kwargs = {}
        if spores_config["skip_cost_op"]:
            print(
                "Skipping cost optimal run and using model_data to initialise SPORES directly"
            )
            return run_plan(model_data, timings, backend, build_only=True, **kwargs)
        else:
            # Run once for the 'cost-optimal' solution
            return run_plan(
                model_data, timings, backend, build_only, persistent=False, **kwargs
            )

    def _initialise_spores_number():
        if "spores" in model_data.coords and spores_config["skip_cost_op"]:
            return model_data.spores.max().item()
        else:
            return 0

    def _error_on_malformed_input():
        if backend_rerun:
            try:
                backend_rerun.obj()
            except ValueError:  # model has not yet been run
                pass
            else:
                raise exceptions.ModelError(
                    "Cannot run SPORES if the backend model already has a solution. "
                    "Consider using the `build_only` optional `run()` argument to avoid this."
                )
        if "spores" in model_data.filter_by_attrs(is_result=0).squeeze().dims:
            raise exceptions.ModelError(
                "Cannot run SPORES with a SPORES dimension in any input (e.g. `cost_energy_cap`)."
            )

    _error_on_malformed_input()

    if backend_rerun:
        model_data = _combine_spores_results_and_inputs(
            backend_rerun, xr.Dataset(), 0, model_data=model_data
        )
    run_config = UpdateObserverDict(
        initial_yaml_string=model_data.attrs["run_config"],
        name="run_config",
        observer=model_data,
    )

    spores_config = run_config["spores_options"]
    if "spores" in model_data.dims and model_data.spores.size == 1:
        model_data = model_data.squeeze("spores")

    init_spores_scores = (
        model_data.cost_energy_cap.loc[{"costs": [spores_config["score_cost_class"]]}]
        .to_series()
        .dropna()
    )
    spores_results = {}

    results, backend_model, opt = _initialise_backend_model()
    if build_only:
        return results, backend_model, opt

    init_spore = _initialise_spores_number()

    if spores_config["skip_cost_op"]:
        cumulative_spores_scores = init_spores_scores.copy()
        _update_to_spores_objective(backend_model)
    elif results.attrs["termination_condition"] in ["optimal", "feasible"]:
        results.attrs["objective_function_value"] = backend_model.obj()
        if spores_config["save_per_spore"] is True:
            _save_spore(backend_model, results, init_spore, model_data=model_data)
        # Storing results and scores in the specific dictionaries
        _add_results_to_list(backend_model, spores_results, results, 0)
        cumulative_spores_scores = init_spores_scores + _cap_loc_score_default(results)
        # Set group constraint "cost_max" equal to slacked cost
        _update_slack_cost_constraint(backend_model)
        _update_to_spores_objective(backend_model)
        # Update "spores_score" based on previous iteration
        _update_spores_score(backend_model, cumulative_spores_scores)

        log_time(
            logger,
            timings,
            "run_solution_returned",
            time_since_run_start=True,
            comment="Backend: generated solution array for the cost-optimal case",
        )
    else:
        _warn_on_infeasibility()
        return results, backend_model, opt

    # Iterate over the number of SPORES requested by the user
    for _spore in range(init_spore + 1, spores_config["spores_number"] + 1):
        print(f"Running SPORES {_spore}")
        if opt is not None and "persistent" in opt.type:
            opt = interface.regenerate_persistent_pyomo_solver(
                backend_model,
                opt,
                obj=True,
                constraints={
                    "cost_investment_constraint": cumulative_spores_scores.index,
                    "cost_constraint": cumulative_spores_scores.index,
                },
            )
        else:
            opt = None
        results, backend_model, opt = run_plan(
            model_data,
            timings,
            backend,
            build_only=False,
            backend_rerun=backend_model,
            allow_warmstart=False,
            persistent=True,
            opt=opt,
        )

        if results.attrs["termination_condition"] in ["optimal", "feasible"]:
            results.attrs["objective_function_value"] = backend_model.obj()
            if spores_config["save_per_spore"] is True:
                _save_spore(backend_model, results, _spore)
            # Storing results and scores in the specific dictionaries
            _add_results_to_list(backend_model, spores_results, results, _spore)
            print(f"Updating capacity scores from {cumulative_spores_scores.sum()}...")
            cumulative_spores_scores += _cap_loc_score_default(results)
            print(f"... to {cumulative_spores_scores.sum()}")
            # Update "spores_score" based on previous iteration
            _update_spores_score(backend_model, cumulative_spores_scores)
        else:
            _warn_on_infeasibility()
            break
        log_time(
            logger,
            timings,
            "run_solution_returned",
            time_since_run_start=True,
            comment=f"Backend: generated solution array for the SPORE {_spore}",
        )

    results = xr.concat(
        spores_results.values(), dim=pd.Index(spores_results.keys(), name="spores")
    )

    return results, backend_model, opt


def run_operate(model_data, timings, backend, build_only):
    """
    For use when mode is 'operate', to allow the model to be built, edited, and
    iteratively run within Pyomo.

    """
    log_time(
        logger,
        timings,
        "run_start",
        comment="Backend: starting model run in operational mode",
    )

    defaults = UpdateObserverDict(
        initial_yaml_string=model_data.attrs["defaults"],
        name="defaults",
        observer=model_data,
    )
    run_config = UpdateObserverDict(
        initial_yaml_string=model_data.attrs["run_config"],
        name="run_config",
        observer=model_data,
    )

    # New param defaults = old maximum param defaults (e.g. energy_cap gets default from energy_cap_max)
    operate_params = {
        k.replace("_max", ""): v for k, v in defaults.items() if k.endswith("_max")
    }
    operate_params[
        "purchased"
    ] = 0  # no _max to work from here, so we hardcode a default

    defaults.update(operate_params)

    # Capacity results (from plan mode) can be used as the input to operate mode
    if any(model_data.filter_by_attrs(is_result=1).data_vars) and run_config.get(
        "operation.use_cap_results", False
    ):
        # Anything with is_result = 1 will be ignored in the Pyomo model
        for varname, varvals in model_data.data_vars.items():
            if varname in operate_params.keys():
                varvals.attrs["is_result"] = 1
                varvals.attrs["operate_param"] = 1

    else:
        cap_max = xr.merge(
            [
                v.rename(k.replace("_max", ""))
                for k, v in model_data.data_vars.items()
                if "_max" in k
            ]
        )
        cap_equals = xr.merge(
            [
                v.rename(k.replace("_equals", ""))
                for k, v in model_data.data_vars.items()
                if "_equals" in k
            ]
        )
        caps = cap_max.update(cap_equals)
        for cap in caps.data_vars.values():
            cap.attrs["is_result"] = 1
            cap.attrs["operate_param"] = 1
        model_data.update(caps)

    comments, warnings, errors = checks.check_operate_params(model_data)
    exceptions.print_warnings_and_raise_errors(warnings=warnings, errors=errors)

    # Initialize our variables
    solver = run_config["solver"]
    solver_io = run_config.get("solver_io", None)
    solver_options = run_config.get("solver_options", None)
    save_logs = run_config.get("save_logs", None)
    window = run_config["operation"]["window"]
    horizon = run_config["operation"]["horizon"]
    window_to_horizon = horizon - window

    # get the cumulative sum of timestep resolution, to find where we hit our window and horizon
    timestep_cumsum = model_data.timestep_resolution.cumsum("timesteps").to_pandas()
    # get the timesteps at which we start and end our windows
    window_ends = timestep_cumsum.where(
        (timestep_cumsum % window == 0) | (timestep_cumsum == timestep_cumsum[-1])
    )
    window_starts = timestep_cumsum.where(
        (~np.isnan(window_ends.shift(1))) | (timestep_cumsum == timestep_cumsum[0])
    ).dropna()

    window_ends = window_ends.dropna()
    horizon_ends = timestep_cumsum[
        timestep_cumsum.isin(window_ends.values + window_to_horizon)
    ]
    # solves bug where horizon_ends is empty if window < # of timesteps < horizon
    if len(horizon_ends) == 0:
        horizon_ends = timestep_cumsum[timestep_cumsum == window_ends.values[-1]]

    if not any(window_starts):
        raise exceptions.ModelError(
            "Not enough timesteps or incorrect timestep resolution to run in "
            "operational mode with an optimisation window of {}".format(window)
        )

    # We will only update timseries parameters
    timeseries_data_vars = [
        k
        for k, v in model_data.data_vars.items()
        if "timesteps" in v.dims and v.attrs["is_result"] == 0
    ]

    # Loop through each window, solve over the horizon length, and add result to
    # result_array we only go as far as the end of the last horizon, which may
    # clip the last bit of data
    result_array = []
    # track whether each iteration finds an optimal solution or not
    terminations = []

    if build_only:
        iterations = [0]
    else:
        iterations = range(len(window_starts))

    for i in iterations:
        start_timestep = window_starts.index[i]

        # Build full model in first instance
        if i == 0:
            warmstart = False
            end_timestep = horizon_ends.index[i]
            timesteps = slice(start_timestep, end_timestep)
            window_model_data = model_data.loc[dict(timesteps=timesteps)]

            log_time(
                logger,
                timings,
                "model_gen_1",
                comment="Backend: generating initial model",
            )

            backend_model = backend.generate_model(window_model_data)

        # Build the full model in the last instance(s),
        # where the number of timesteps may be less than the horizon length
        elif i > len(horizon_ends) - 1 or i == iterations[-1]:
            warmstart = False
            end_timestep = window_ends.index[i]
            timesteps = slice(start_timestep, end_timestep)
            window_model_data = model_data.loc[dict(timesteps=timesteps)]

            log_time(
                logger,
                timings,
                "model_gen_{}".format(i + 1),
                comment=(
                    "Backend: iteration {}: generating new model for "
                    "end of timeseries, with horizon = {} timesteps".format(
                        i + 1, window_ends[i] - window_starts[i]
                    )
                ),
            )

            backend_model = backend.generate_model(window_model_data)

        # Update relevent Pyomo Params in intermediate instances
        else:
            warmstart = True
            end_timestep = horizon_ends.index[i]
            timesteps = slice(start_timestep, end_timestep)
            window_model_data = model_data.loc[dict(timesteps=timesteps)]

            log_time(
                logger,
                timings,
                "model_gen_{}".format(i + 1),
                comment="Backend: iteration {}: updating model parameters".format(
                    i + 1
                ),
            )
            # Pyomo model sees the same timestamps each time, we just change the
            # values associated with those timestamps
            for var in timeseries_data_vars:
                # New values
                var_series = (
                    window_model_data[var].to_series().dropna().replace("inf", np.inf)
                )
                # Same timestamps
                var_series.index = backend_model.__calliope_model_data["data"][
                    var
                ].keys()
                var_dict = var_series.to_dict()
                # Update pyomo Param with new dictionary

                getattr(backend_model, var).store_values(var_dict)

        if not build_only:
            log_time(
                logger,
                timings,
                "model_run_{}".format(i + 1),
                time_since_run_start=True,
                comment="Backend: iteration {}: sending model to solver".format(i + 1),
            )
            # After iteration 1, warmstart = True, which should speed up the process
            # Note: Warmstart isn't possible with GLPK (dealt with later on)
            _results, _opt = backend.solve_model(
                backend_model,
                solver=solver,
                solver_io=solver_io,
                solver_options=solver_options,
                save_logs=save_logs,
                warmstart=warmstart,
            )

            log_time(
                logger,
                timings,
                "run_solver_exit_{}".format(i + 1),
                time_since_run_start=True,
                comment="Backend: iteration {}: solver finished running".format(i + 1),
            )
            # xarray dataset is built for each iteration
            _termination = backend.load_results(backend_model, _results, _opt)
            terminations.append(_termination)
            if _termination in ["optimal", "feasible"]:
                _results = backend.get_result_array(backend_model, model_data)
            else:
                _results = xr.Dataset()

            # We give back the actual timesteps for this iteration and take a slice
            # equal to the window length
            _results["timesteps"] = window_model_data.timesteps.copy()

            # We always save the window data. Until the last window(s) this will crop
            # the window_to_horizon timesteps. In the last window(s), optimistion will
            # only be occurring over a window length anyway
            _results = _results.loc[dict(timesteps=slice(None, window_ends.index[i]))]
            result_array.append(_results)

            # Set up initial storage for the next iteration
            if "loc_techs_store" in model_data.dims.keys() and _termination in ["optimal", "feasible"]:
                storage_initial = (
                    _results.storage.loc[{"timesteps": window_ends.index[i]}].drop_vars(
                        "timesteps"
                    )
                    / model_data.storage_cap
                )
                model_data["storage_initial"].loc[
                    storage_initial.coords
                ] = storage_initial.values
                backend_model.storage_initial.store_values(
                    storage_initial.to_series().dropna().to_dict()
                )

            # Set up total operated units for the next iteration
            if "loc_techs_milp" in model_data.dims.keys() and _termination in ["optimal", "feasible"]:
                operated_units = _results.operating_units.sum("timesteps").astype(int)
                model_data["operated_units"].loc[{}] += operated_units.values
                backend_model.operated_units.store_values(
                    operated_units.to_series().dropna().to_dict()
                )

            log_time(
                logger,
                timings,
                "run_solver_exit_{}".format(i + 1),
                time_since_run_start=True,
                comment="Backend: iteration {}: generated solution array".format(i + 1),
            )

    if build_only:
        results = xr.Dataset()
        _opt = None
    else:
        # Concatenate results over the timestep dimension to get a single
        # xarray Dataset of interest
        results = xr.concat(result_array, dim="timesteps")
        if all(i == "optimal" for i in terminations):
            results.attrs["termination_condition"] = "optimal"
        elif all(i in ["optimal", "feasible"] for i in terminations):
            results.attrs["termination_condition"] = "feasible"
        else:
            results.attrs["termination_condition"] = ",".join(terminations)

        log_time(
            logger,
            timings,
            "run_solution_returned",
            time_since_run_start=True,
            comment="Backend: generated full solution array",
        )

    return results, backend_model, _opt



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

import logging
import os
from contextlib import redirect_stdout, redirect_stderr

import numpy as np
import pandas as pd
import xarray as xr

import pyomo.core as po  # pylint: disable=import-error
from pyomo.opt import SolverFactory  # pylint: disable=import-error

# pyomo.environ is needed for pyomo solver plugins
import pyomo.environ as pe  # pylint: disable=unused-import,import-error

# TempfileManager is required to set log directory
from pyomo.common.tempfiles import TempfileManager  # pylint: disable=import-error

from calliope.backend.pyomo.util import (
    get_var,
    get_domain,
    datetime_to_string,
    string_to_datetime,
)
from calliope.backend.pyomo import constraints
from calliope.core.util.tools import load_function
from calliope.core.util.logging import LogWriter
from calliope.core.util.dataset import reorganise_xarray_dimensions
from calliope import exceptions
from calliope.core.attrdict import AttrDict

logger = logging.getLogger(__name__)


def generate_model(model_data):
    """
    Generate a Pyomo model.

    """

    backend_model = po.ConcreteModel()
    model_data = datetime_to_string(backend_model, model_data)

    logger.info("Loading sets")
    # Sets
    for coord_name, coord_data in model_data.coords.items():
        if not coord_data.shape:
            continue
        set_data = list(coord_data.data)
        # Ensure that time steps are pandas.Timestamp objects
        if isinstance(set_data[0], np.datetime64):
            set_data = pd.to_datetime(set_data)
        setattr(backend_model, coord_name, po.Set(initialize=set_data, ordered=True))
    logger.info("Loading parameters")
    # "Parameters"
    model_data_dict = {
        "data": {
            k: v.to_series().dropna().replace("inf", np.inf).to_dict()
            for k, v in model_data.data_vars.items()
            if v.attrs["is_result"] == 0 or v.attrs.get("operate_param", 0) == 1
        },
        "dims": {
            k: v.dims
            for k, v in model_data.data_vars.items()
            if v.attrs["is_result"] == 0 or v.attrs.get("operate_param", 0) == 1
        },
        "sets": list(model_data.coords),
        "attrs": {k: v for k, v in model_data.attrs.items() if k != "defaults"},
    }

    # Dims in the dict's keys are ordered as in model_data, which is enforced
    # in model_data generation such that timesteps are always last and the
    # remainder of dims are in alphabetic order
    backend_model.__calliope_model_data = model_data_dict
    backend_model.__calliope_defaults = AttrDict.from_yaml_string(
        model_data.attrs["defaults"]
    )
    backend_model.__calliope_run_config = AttrDict.from_yaml_string(
        model_data.attrs["run_config"]
    )

    for k, v in model_data_dict["data"].items():
        _kwargs = {
            "initialize": v,
            "mutable": True,
            "within": getattr(po, get_domain(model_data[k])),
        }
        if not pd.isnull(backend_model.__calliope_defaults.get(k, None)):
            _kwargs["default"] = backend_model.__calliope_defaults[k]
        # In operate mode, e.g. energy_cap is a parameter, not a decision variable,
        # so add those in.
        if (
            backend_model.__calliope_run_config["mode"] == "operate"
            and model_data[k].attrs.get("operate_param") == 1
        ):
            dims = [getattr(backend_model, model_data_dict["dims"][k][0])]
        else:
            dims = [getattr(backend_model, i) for i in model_data_dict["dims"][k]]
        setattr(backend_model, k, po.Param(*dims, **_kwargs))

    for option_name, option_val in backend_model.__calliope_run_config[
        "objective_options"
    ].items():
        if option_name == "cost_class":
            objective_cost_class = {
                k: v for k, v in option_val.items() if k in backend_model.costs
            }

            backend_model.objective_cost_class = po.Param(
                backend_model.costs,
                initialize=objective_cost_class,
                mutable=True,
                within=po.Reals,
            )
        else:
            setattr(backend_model, "objective_" + option_name, option_val)

    # Variables
    load_function("calliope.backend.pyomo.variables.initialize_decision_variables")(
        backend_model
    )

    # Constraints
    constraints_to_add = [
        i.split(".py")[0]
        for i in os.listdir(constraints.__path__[0])
        if not i.startswith("_") and not i.startswith(".")
    ]

    # The list is sorted to ensure that some constraints are added after pyomo
    # expressions have been created in other constraint files.
    # Ordering is given by the number assigned to the variable ORDER within each
    # file (higher number = added later).
    try:
        constraints_to_add.sort(
            key=lambda x: load_function(
                "calliope.backend.pyomo.constraints." + x + ".ORDER"
            )
        )
    except AttributeError as e:
        raise AttributeError(
            "{}. This attribute must be set to an integer value based "
            "on the order in which the constraints in the file {}.py should be "
            "loaded relative to constraints in other constraint files. If order "
            "does not matter, set ORDER to a value of 10.".format(
                e.args[0], e.args[0].split(".")[-1].split("'")[0]
            )
        )

    logger.info(
        "constraints are loaded in the following order: {}".format(constraints_to_add)
    )

    for c in constraints_to_add:
        logger.info(f"creating {c} constraints")
        load_function("calliope.backend.pyomo.constraints." + c + ".load_constraints")(
            backend_model
        )

    # FIXME: Optional constraints
    # optional_constraints = model_data.attrs['constraints']
    # if optional_constraints:
    #     for c in optional_constraints:
    #         self.add_constraint(load_function(c))

    # Objective function
    # FIXME re-enable loading custom objectives

    # fetch objective function by name, pass through objective options
    # if they are present
    objective_function = (
        "calliope.backend.pyomo.objective."
        + backend_model.__calliope_run_config["objective"]
    )
    load_function(objective_function)(backend_model)

    model_data = string_to_datetime(backend_model, model_data)

    return backend_model


def solve_model(
    backend_model,
    solver,
    solver_io=None,
    solver_options=None,
    save_logs=False,
    opt=None,
    **solve_kwargs,
):
    """
    Solve a Pyomo model using the chosen solver and all necessary solver options

    Returns a Pyomo results object
    """
    if opt is None:
        opt = SolverFactory(solver, solver_io=solver_io)
        if "persistent" in solver:
            solve_kwargs.update({"save_results": False, "load_solutions": False})
            opt.set_instance(backend_model)

    if solver_options:
        for k, v in solver_options.items():
            opt.options[k] = v

    if save_logs:
        solve_kwargs.update({"symbolic_solver_labels": True, "keepfiles": True})
        os.makedirs(save_logs, exist_ok=True)
        TempfileManager.tempdir = save_logs  # Sets log output dir
    if "warmstart" in solve_kwargs.keys() and solver in ["glpk", "cbc"]:
        if solve_kwargs.pop("warmstart") is True:
            exceptions.warn(
                "The chosen solver, {}, does not suport warmstart, which may "
                "impact performance.".format(solver)
            )

    with redirect_stdout(LogWriter(logger, "debug", strip=True)):
        with redirect_stderr(LogWriter(logger, "error", strip=True)):
            # Ignore most of gurobipy's logging, as it's output is
            # already captured through STDOUT
            logging.getLogger("gurobipy").setLevel(logging.ERROR)
            if "persistent" in solver:
                results = opt.solve(tee=True, **solve_kwargs)
            else:
                results = opt.solve(backend_model, tee=True, **solve_kwargs)
    return results, opt


def load_results(backend_model, results, opt):
    """Load results into model instance for access via model variables."""
    termination = results.solver.termination_condition

    if termination == pe.TerminationCondition.optimal:
        try:
            opt.load_vars()
            this_result = True
        except AttributeError:
            this_result = backend_model.solutions.load_from(results)

    if termination != pe.TerminationCondition.optimal or this_result is False:
        logger.critical("Problem status:")
        for l in str(results.Problem).split("\n"):
            logger.critical(l)
        logger.critical("Solver status:")
        for l in str(results.Solver).split("\n"):
            logger.critical(l)

        if termination != pe.TerminationCondition.optimal:
            message = "Model solution was non-optimal."
        else:
            message = "Could not load results into model instance."

        exceptions.BackendWarning(message)

    return str(termination)


def get_result_array(backend_model, model_data):
    """
    From a Pyomo model object, extract decision variable data and return it as
    an xarray Dataset. Any rogue input parameters that are constructed inside
    the backend (instead of being passed by calliope.Model().inputs) are also
    added to calliope.Model()._model_data in-place.
    """
    all_variables = {
        i.name: get_var(backend_model, i.name)
        for i in backend_model.component_objects()
        if isinstance(i, po.base.Var)
    }
    all_variables.update(
        {
            i.name: get_var(backend_model, i.name, expr=True)
            for i in backend_model.component_objects()
            if isinstance(i, po.base.Expression)
        }
    )
    # Get any parameters that did not appear in the user's model.inputs Dataset
    all_params = {
        i.name: get_var(backend_model, i.name, expr=True)
        for i in backend_model.component_objects(ctype=po.base.param.IndexedParam)
        if i.name not in model_data.data_vars.keys() and "objective_" not in i.name
    }

    results = string_to_datetime(
        backend_model, reorganise_xarray_dimensions(xr.Dataset(all_variables))
    )

    if all_params:
        additional_inputs = reorganise_xarray_dimensions(xr.Dataset(all_params))
        for var in additional_inputs.data_vars:
            additional_inputs[var].attrs["is_result"] = 0
        model_data.update(additional_inputs)
        model_data = string_to_datetime(backend_model, model_data)

    return results



import pytest

import calliope
from calliope.test.common.util import build_test_model as build_model


class TestFutureWarnings:
    def test_group_share_warning(self):
        override = {
            "model.group_share.test_supply_elec.carrier_prod_min.electricity": 0.85
        }
        expected_warning_msg = "`group_share` constraints will be removed in v0.7.0"
        with pytest.warns(FutureWarning, match=expected_warning_msg):
            build_model(override_dict=override, scenario="simple_supply,one_day")

    @pytest.mark.filterwarnings("always::FutureWarning")
    def test_default_cost_class_warning(self):
        expected_warning_msg = (
            "There will be no default cost class for the objective function in v0.7.0"
        )
        with pytest.warns(FutureWarning, match=expected_warning_msg):
            build_model(scenario="simple_supply,one_day")

    @pytest.mark.filterwarnings(
        "error:.*There will be no default cost class for the objective function in v0.7.0.*:FutureWarning"
    )
    def test_default_cost_class_from_override_no_warning(self):
        override = {"run.objective_options.cost_class": {"monetary": 1}}
        build_model(override_dict=override, scenario="simple_supply,one_day")

    @pytest.mark.filterwarnings(
        "error:.*There will be no default cost class for the objective function in v0.7.0.*:FutureWarning"
    )
    def test_default_cost_class_from_scenario_no_warning(self):
        build_model(scenario="simple_supply,one_day,explicit_cost_class")

    @pytest.mark.filterwarnings("always::FutureWarning")
    def test_future_warning_for_charge_rate(self):
        with pytest.warns(
            FutureWarning,
            match="`charge_rate` is renamed to `energy_cap_per_storage_cap_max`",
        ):
            calliope.examples.national_scale(
                override_dict={"techs.battery.constraints.charge_rate": 5}
            )



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

logger.py
~~~~~~~~~~

Create the Calliope logger object and apply other logging tools/functionality

"""

import datetime
import logging
import sys

_time_format = "%Y-%m-%d %H:%M:%S"


def setup_root_logger(verbosity, capture_warnings):
    root_logger = logging.getLogger()  # Get the root logger

    # Remove any existing output handlers from root logger
    if root_logger.hasHandlers():
        for handler in root_logger.handlers:
            root_logger.removeHandler(handler)

    # Create a console log handler with decent formatting and attach it
    formatter = logging.Formatter(
        "[%(asctime)s] %(levelname)-8s %(message)s", datefmt=_time_format
    )
    console = logging.StreamHandler(stream=sys.stdout)
    console.setFormatter(formatter)
    root_logger.addHandler(console)
    root_logger.setLevel(verbosity.upper())

    if capture_warnings:
        logging.captureWarnings(True)
        pywarning_logger = logging.getLogger("py.warnings")
        pywarning_logger.setLevel(verbosity.upper())

    return root_logger


def set_log_verbosity(
    verbosity="info", include_solver_output=True, capture_warnings=True
):
    """
    Set the verbosity of logging and setup the root logger to log to
    console (stdout) with timestamp output formatting.

    Parameters
    ----------
    verbosity : str, default 'info'
        Logging level to display across all of Calliope. Can be one of
        'debug', 'info', 'warning', 'error', or 'critical'.
    include_solver_output : bool, default True
        If True, the logging level for just the backend model is set to
        DEBUG, which turns on display of solver output.
    capture_warnings : bool, default True
        If True, also capture all warnings and log them to the WARNING
        level. This results in more consistent output when running
        interactively.

    """
    backend_logger = logging.getLogger("calliope.backend.pyomo.model")
    if include_solver_output is True:
        backend_logger.setLevel("DEBUG")
    else:
        backend_logger.setLevel(verbosity.upper())

    setup_root_logger(verbosity=verbosity, capture_warnings=capture_warnings)


def log_time(
    logger, timings, identifier, comment=None, level="info", time_since_run_start=False
):
    if comment is None:
        comment = identifier

    timings[identifier] = now = datetime.datetime.now()

    if time_since_run_start and "run_start" in timings:
        time_diff = now - timings["run_start"]
        comment += ". Time since start of model run: {}".format(time_diff)

    getattr(logger, level)(comment)


class LogWriter:
    def __init__(self, logger, level, strip=False):
        self.logger = logger
        self.level = level
        self.strip = strip

    def write(self, message):
        if message != "\n":
            if self.strip:
                message = message.strip()
            getattr(self.logger, self.level)(message)

    def flush(self):
        pass



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

from calliope import exceptions
import xarray as xr
import pandas as pd


def get_loc_techs(loc_techs, tech=None, loc=None):
    """
    Get a list of loc_techs associated with the given technology and/or location.
    If multiple of both loc and tech are given, the function will return any
    combination of members of loc and tech lists found in loc_techs.

    Parameters
    ----------
    loc_techs : list
        set of loc_techs to search for the relevant tech and/or loc
    tech : string or list of strings, default None
        technology/technologies to search for in the set of location:technology
    loc : string or list of strings, default None
        location(s) to search for in the set of location:technology

    Returns
    -------
    relevant_loc_techs : list of strings

    """
    # If both are strings, there is only one loc:tech possibility to look for
    if (
        isinstance(tech, str)
        and isinstance(loc, str)
        and "::".join([loc, tech]) in loc_techs
    ):
        relevant_loc_techs = ["::".join([loc, tech])]

    tech = [tech] if tech is not None and isinstance(tech, str) else tech
    loc = [loc] if loc is not None and isinstance(loc, str) else loc

    if tech and not loc:
        relevant_loc_techs = [i for i in loc_techs if i.split("::")[1] in tech]
    elif loc and not tech:
        relevant_loc_techs = [i for i in loc_techs if i.split("::")[0] in loc]
    elif loc and tech:
        loc_techs_set = set(tuple(i.split("::")) for i in loc_techs)
        possible_loc_techs = set((l, t) for l in loc for t in tech)
        relevant_loc_techs = [
            "::".join(i) for i in possible_loc_techs.intersection(loc_techs_set)
        ]
    else:
        relevant_loc_techs = [None]

    return relevant_loc_techs


def reorganise_xarray_dimensions(data):
    """
    Reorganise Dataset or DataArray dimensions to be alphabetical *except*
    `timesteps`, which must always come last in any DataArray's dimensions
    """

    if not (isinstance(data, xr.Dataset) or isinstance(data, xr.DataArray)):
        raise TypeError(
            "Must provide either xarray Dataset or DataArray to be reorganised"
        )

    steps = [i for i in ["datesteps", "timesteps"] if i in data.dims]

    if isinstance(data, xr.Dataset):
        new_dims = (sorted(list(set(data.dims.keys()) - set(steps)))) + steps
    elif isinstance(data, xr.DataArray):
        new_dims = (sorted(list(set(data.dims) - set(steps)))) + steps

    updated_data = data.transpose(*new_dims).reindex({k: data[k] for k in new_dims})

    return updated_data


def split_loc_techs(data_var, return_as="DataArray"):
    """
    Get a DataArray with locations technologies, and possibly carriers
    split into separate coordinates.

    Parameters
    ----------
    data_var : xarray DataArray
        Variable from Calliope model_data, to split loc_techs dimension
    return_as : string
        'DataArray' to return xarray DataArray, 'MultiIndex DataArray' to return
        xarray DataArray with loc_techs as a MultiIndex,
        or 'Series' to return pandas Series with dimensions as a MultiIndex

    Returns
    -------
    updated_data_var : xarray DataArray of pandas Series
    """

    # Separately find the loc_techs(_carriers) dimension and all other dimensions
    loc_tech_dim = [i for i in data_var.dims if "loc_tech" in i]
    if not loc_tech_dim:
        loc_tech_dim = [i for i in data_var.dims if "loc_carrier" in i]

    if not loc_tech_dim:
        if return_as == "Series":
            return data_var.to_series()
        elif return_as in ["DataArray", "MultiIndex DataArray"]:
            return data_var
        else:
            raise ValueError(
                "`return_as` must be `DataArray`, `Series`, or "
                "`MultiIndex DataArray`, but `{}` given".format(return_as)
            )

    elif len(loc_tech_dim) > 1:
        e = exceptions.ModelError
        raise e(
            "Cannot split loc_techs or loc_tech_carriers dimension "
            "for DataArray {}".format(data_var.name)
        )

    loc_tech_dim = loc_tech_dim[0]
    # xr.Datarray -> pd.Series allows for string operations
    data_var_idx = data_var[loc_tech_dim].to_index()
    index_list = data_var_idx.str.split("::").tolist()

    # carrier_prod, carrier_con, and carrier_export will return an index_list
    # of size 3, all others will be an index list of size 2
    possible_names = ["loc", "tech", "carrier"]
    names = [i + "s" for i in possible_names if i in loc_tech_dim]

    data_var_midx = pd.MultiIndex.from_tuples(index_list, names=names)

    # Replace the Datarray loc_tech_dim with this new MultiIndex
    updated_data_var = data_var.copy()
    updated_data_var.coords[loc_tech_dim] = data_var_midx

    if return_as == "MultiIndex DataArray":
        return updated_data_var

    elif return_as == "Series":
        return reorganise_xarray_dimensions(updated_data_var.unstack()).to_series()

    elif return_as == "DataArray":
        return reorganise_xarray_dimensions(updated_data_var.unstack())

    else:
        raise ValueError(
            "`return_as` must be `DataArray`, `Series`, or "
            "`MultiIndex DataArray`, but `{}` given".format(return_as)
        )



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

masks.py
~~~~~~~~

Functions to pick timesteps from data given certain criteria.

"""

import pandas as pd

from calliope.time import funcs
from calliope.core.util.dataset import split_loc_techs
from calliope import exceptions


def _get_array(data, var, tech, **kwargs):
    subset = {"techs": tech}
    if kwargs is not None:
        subset.update({k: v for k, v in kwargs.items()})

    unusable_dims = (
        set(subset.keys()).difference(["techs", "locs"]).difference(data[var].dims)
    )
    if unusable_dims:
        raise exceptions.ModelError(
            "Attempting to mask time based on  technology {}, "
            "but dimension(s) {} do not exist for parameter {}".format(
                tech, unusable_dims, var.name
            )
        )

    arr = split_loc_techs(data[var].copy()).loc[subset]
    arr = arr.mean(dim=[i for i in arr.dims if i != "timesteps"]).to_pandas()
    return arr


def zero(data, tech, var="resource", **kwargs):
    """
    Returns timesteps where ``var`` for the technology ``tech`` is zero.

    kwargs are additional dimensions to subset on, for example,
    ``locs=['location1', 'location2]``

    """
    s = _get_array(data, var, tech, **kwargs)

    return s[s == 0].index


def _concat_indices(indices):
    return pd.concat([i.to_series() for i in indices]).sort_index().index


def _get_minmax_timestamps(series, length, n, how="max", padding=None):
    # Get the max/min timestamps
    group = series.groupby(pd.Grouper(freq=length)).mean()
    timesteps = []
    for _ in range(n):
        if how == "max":
            ts = group.idxmax()
        elif how == "min":
            ts = group.idxmin()
        timesteps.append(ts)
        group = group.drop(ts)

    # Get range of timestamps including padding
    full_timesteps = []
    for ts in timesteps:
        ts_end = ts + pd.Timedelta(length)
        if padding is not None:
            ts -= pd.Timedelta(padding)
            ts_end += pd.Timedelta(padding)
        ts_range = series[ts:ts_end].index[:-1]
        full_timesteps.append(ts_range)

    ts_index = _concat_indices(full_timesteps)

    return ts_index


def extreme(
    data,
    tech,
    var="resource",
    how="max",
    length="1D",
    n=1,
    groupby_length=None,
    padding=None,
    normalize=True,
    **kwargs,
):
    """
    Returns timesteps for period of ``length`` where ``var`` for the technology
    ``tech`` across the given list of ``locations`` is either minimal
    or maximal.

    Parameters
    ----------
    data : xarray.Dataset
    tech : str
        Technology whose `var` to find extreme for.
    var : str, optional
        default 'resource'
    how : str, optional
        'max' (default) or 'min'.
    length : str, optional
        Defaults to '1D'.
    n : int, optional
        Number of periods of `length` to look for, default is 1.
    groupby_length : str, optional
        Group time series and return `n` periods of `length`
        for each group.
    padding : str, optional
        Either Pandas frequency (e.g. '1D') or 'calendar_week'.
        If Pandas frequency, symmetric padding is undertaken, either side of `length`
        If 'calendar_week', padding is fit to the calendar week in which the
        extreme day(s) are found.
    normalize : bool, optional
        If True (default), data is normalized
        using :func:`~calliope.time.funcs.normalized_copy`.
    kwargs : dict, optional
        Dimensions of the selected var over which to index. Any remaining
        dimensions will be flattened by mean

    """
    if normalize:
        # Only normalise the desired var as rest of data may contain
        # non-numeric variables!
        data_n = funcs.normalized_copy(data[var].to_dataset(name=var))
    else:
        data_n = data
    arr = _get_array(data_n, var, tech, **kwargs)
    return _extreme_with_padding(arr, how, length, n, groupby_length, padding)


def extreme_diff(
    data,
    tech0,
    tech1,
    var="resource",
    how="max",
    length="1D",
    n=1,
    groupby_length=None,
    padding=None,
    normalize=True,
    **kwargs,
):
    """
    Returns timesteps for period of ``length`` where the diffence in extreme
    value for ``var`` between technologies ``tech0`` and ``tech1`` is either a
    minimum or a maximum.

    Parameters
    ----------
    data : xarray.Dataset
    tech0 : str
        First technology for which we find the extreme of `var`
    tech1 : str
        Second technology for which we find the extreme of `var`
    var : str, optional
        default 'resource'
    how : str, optional
        'max' (default) or 'min'.
    length : str, optional
        Defaults to '1D'.
    n : int, optional
        Number of periods of `length` to look for, default is 1.
    groupby_length : str, optional
        Group time series and return `n` periods of `length`
        for each group.
    padding : str, optional
        Either Pandas frequency (e.g. '1D') or 'calendar_week'.
        If Pandas frequency, symmetric padding is undertaken, either side of `length`
        If 'calendar_week', padding is fit to the calendar week in which the
        extreme day(s) are found.
    normalize : bool, optional
        If True (default), data is normalized
        using :func:`~calliope.time.funcs.normalized_copy`.
    kwargs : dict, optional
        Dimensions of the selected var over which to index. Any remaining
        dimensions will be flattened by mean

    """
    if normalize:
        # Only normalise the desired var as rest of data may contain
        # non-numeric variables!
        data_n = funcs.normalized_copy(data[var].to_dataset(name=var))
    else:
        data_n = data
    arr0 = _get_array(data_n, var, tech0, **kwargs)
    arr1 = _get_array(data_n, var, tech1, **kwargs)
    arr = arr0 - arr1

    return _extreme_with_padding(arr, how, length, n, groupby_length, padding)


def _extreme(arr, how="max", length="1D", n=1, groupby_length=None, padding=None):

    if groupby_length:
        groupby = pd.Grouper(freq=groupby_length)
        group_indices = []
        grouping = arr.groupby(groupby)
        for k in grouping.groups.keys():
            s = grouping.get_group(k)
            group_indices.append(_get_minmax_timestamps(s, length, n, how, padding))
        ts_index = _concat_indices(group_indices)
    else:
        ts_index = _get_minmax_timestamps(arr, length, n, how, padding)

    return ts_index


def _extreme_with_padding(arr, how, length, n, groupby_length, padding):
    if padding == "calendar_week":
        if n != 1 or length != "1D":
            raise ValueError(
                "calendar_week padding only supports n=1 and length=1D for now."
            )
        result = _extreme(arr, how, length, n, groupby_length, padding=None)
        # get week padding for each day in result
        days = list(result.groupby(result.dayofyear).values())
        weeks = pd.DatetimeIndex(days[0])
        for d in days:
            weeks = weeks.union(_calendar_week_padding(d, arr))
        # concatenate the weeks into one index and drop possible duplicates
        return pd.DatetimeIndex(weeks).drop_duplicates()
    else:
        return _extreme(arr, how, length, n, groupby_length, padding)


def _calendar_week_padding(day, arr):
    """
    Given a day, returns the whole calendar week which contains that day

    """
    days = len(day.day.unique())
    if not days == 1:
        raise ValueError(
            "Only a single day at a time may be used for calendar_week padding, "
            "but {} days were passed.".format(days)
        )

    # Using day of week, figure out how many days before and after to get
    # a complete week
    days_before = 6 - day[0].dayofweek
    days_after = 6 - days_before

    # Turn it into a week
    start_time = day[0] - pd.Timedelta("{}D".format(days_before))
    end_time = day[-1] + pd.Timedelta("{}D".format(days_after))
    before = arr[start_time : day[0]].index[:-1]
    after = arr[day[-1] : end_time].index[1:]
    result_week = before.append(day).append(after)

    return result_week



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

capacity.py
~~~~~~~~~~~

Plot capacity data.

"""

from itertools import product

import plotly.graph_objs as go
from natsort import natsorted

from calliope.postprocess.util import subset_sum_squeeze
from calliope.postprocess.plotting.util import get_data_layout, break_name


def _get_relevant_vars(dataset, array):
    allowed_input_vars = [
        i + j
        for i, j in product(
            ["resource_area", "energy_cap", "resource_cap", "storage_cap", "units"],
            ["_max", "_min", "_equals"],
        )
    ]
    allowed_result_vars = [
        "results",
        "inputs",
        "all",
        "resource_area",
        "energy_cap",
        "resource_cap",
        "storage_cap",
        "units",
        "systemwide_levelised_cost",
        "systemwide_capacity_factor",
    ]

    if (
        isinstance(array, list)
        and not set(array) != set(allowed_input_vars + allowed_result_vars)
    ) or (
        isinstance(array, str) and array not in allowed_input_vars + allowed_result_vars
    ):
        raise ValueError(
            "Cannot plot array={}. as one or more of the elements is not considered "
            "to be a capacity".format(array)
        )

    # relevant_vars are all variables relevant to this plotting instance
    if array == "results":
        relevant_vars = natsorted(allowed_result_vars)
    elif array == "inputs":
        relevant_vars = natsorted(allowed_input_vars)
    elif array == "all":
        relevant_vars = natsorted(allowed_result_vars + allowed_input_vars)
    elif isinstance(array, list):
        relevant_vars = array
    elif isinstance(array, str):
        relevant_vars = [array]

    relevant_vars = [i for i in relevant_vars if i in dataset]

    # Remove all vars that don't actually turn up in the dataset, which is relevant
    # ony really for results vars
    return sorted(list(set(relevant_vars).intersection(dataset.data_vars.keys())))


def _get_var_data(
    cap, model, dataset, visible, subset, sum_dims, squeeze, locations, orientation
):
    if "systemwide" in cap:
        array_cap = subset_sum_squeeze(dataset[cap], subset)
        if "costs" in array_cap.dims and len(array_cap["costs"]) == 1:
            array_cap = array_cap.squeeze("costs")
        elif "costs" in array_cap.dims and len(array_cap["costs"]) > 1:
            raise ValueError(
                "Cannot plot {} without subsetting to pick one cost type "
                "of interest".format(cap)
            )
        if "carriers" not in subset.keys():
            array_cap = array_cap.sortby("carriers")

    else:
        array_cap = model.get_formatted_array(cap).reindex(locs=locations)
        array_cap = subset_sum_squeeze(array_cap, subset, sum_dims, squeeze)

    if len(array_cap.dims) > 2:
        raise ValueError(
            "Maximum two dimensions allowed for plotting capacity, but {} "
            "given as dimensions for {}".format(array_cap.dims, cap)
        )

    if "techs" not in array_cap.dims:
        raise ValueError("Cannot plot capacity without `techs` in dimensions")

    elif "techs" not in subset.keys():
        array_cap = array_cap.sortby("techs")

    data = []

    for tech in array_cap.techs.values:
        if tech not in dataset.techs.values:
            continue

        if (
            "techs_transmission" in dataset
            and tech in dataset.techs_transmission.values
        ):
            continue
        else:
            base_tech = dataset.inheritance.loc[{"techs": tech}].item().split(".")[0]

        if base_tech in "demand":
            continue

        if array_cap.loc[{"techs": tech}].sum() > 0:
            x = array_cap.loc[{"techs": tech}].values
            name = break_name(model._model_data.names.loc[{"techs": tech}].item(), 30)
            if "systemwide" in cap:
                y = natsorted(array_cap.carriers.values)
            else:
                if "locs" in array_cap.dims:
                    y = natsorted(array_cap.locs.values)
                else:  # Single location
                    y = [array_cap.locs.values]

            if orientation == "v":
                x, y = y, x  # Flip axes
                hoverinfo = "y+name"
            else:
                hoverinfo = "x+name"
                x, y = x[::-1], y[::-1]  # Make sure that sorting is from bottom down

            data.append(
                go.Bar(
                    x=x,
                    y=y,
                    visible=visible,
                    name=name,
                    legendgroup=tech,
                    text=tech,
                    hoverinfo=hoverinfo,
                    marker=dict(
                        color=model._model_data.colors.loc[{"techs": tech}].item()
                    ),
                    orientation=orientation,
                )
            )

    return data


def _get_var_layout(cap, dataset, location_axis, value_axis):
    args = {}
    if "area" in cap:
        value_axis_title = "Installed area"
    elif "units" in cap:
        value_axis_title = "Installed units"
    elif "storage" in cap:
        value_axis_title = "Installed storage capacity"
    elif "energy" in cap:
        value_axis_title = "Installed energy capacity"
    elif "systemwide" in cap:
        value_axis_title = cap.replace("_", " ").capitalize()
        args.update({location_axis: {"title": "Carrier"}})
    else:
        value_axis_title = "Installed capacity"

    if "_max" in cap:
        title = value_axis_title.replace("Installed", "Maximum allowed")
    elif "_min" in cap:
        title = value_axis_title.replace("Installed", "Minimum allowed")
    elif "_equal" in cap:
        title = value_axis_title.replace("Installed", "Allowed")
    else:
        title = value_axis_title

    if "systemwide" not in cap:
        args.update({location_axis: {"title": "Location"}})

    # Grouped, not stacked, barcharts for the systemwide variables
    if "systemwide" in cap:
        args["barmode"] = "group"
    else:
        args["barmode"] = "relative"

    args.update({value_axis: dict(title=value_axis_title), "title": title})

    return args


def plot_capacity(
    model, orient="h", array="all", subset={}, sum_dims=None, squeeze=True, **kwargs
):
    """
    Parameters
    ----------
    array : str or list; default = 'all'
        options: 'all', 'results', 'inputs', the name/list of any energy capacity
        DataArray(s) from inputs/results.
        User can specify 'all' for all input/results capacities, 'inputs'
        for just input capacities, 'results' for just results capacities, or the
        name(s) of any data array(s) to plot (in either inputs or results).
        In all but the last case, arrays can be picked from dropdown in visualisation.
        In the last case, output can be saved to SVG.
    orient : str, optional
        'h' for horizontal or 'v' for vertical barchart
    subset : dict, optional
        Dictionary by which data is selected (using xarray indexing `loc[]`).
        Keys any of ['timeseries', 'locs', 'techs', 'carriers', 'costs']).
    sum_dims : str, optional
        List of dimension names to sum plot variable over.
    squeeze : bool, optional
        Whether to squeeze out dimensions containing only single values.

    """
    dataset = model._model_data.copy()
    locations = natsorted(list(dataset.locs.values))

    if orient in ["horizontal", "h"]:
        orientation = "h"
        location_axis = "yaxis"
        value_axis = "xaxis"
    elif orient in ["vertical", "v"]:
        orientation = "v"
        location_axis = "xaxis"
        value_axis = "yaxis"
    else:
        raise ValueError("Orient must be `v`/`vertical` or `h`/`horizontal`")

    layout = {
        location_axis: dict(
            title="Location",
            showticklabels=True,  # FIXME: Ensure labels do not get hidden if little vertical space
        ),
        "legend": (dict(traceorder="reversed")),
        "autosize": True,
        "hovermode": "closest",
    }

    relevant_vars = _get_relevant_vars(dataset, array)

    data, layout = get_data_layout(
        _get_var_data,
        _get_var_layout,
        relevant_vars,
        layout,
        model,
        dataset,
        subset,
        sum_dims,
        squeeze,
        get_var_data_kwargs={"locations": locations, "orientation": orientation},
        get_var_layout_kwargs={
            "location_axis": location_axis,
            "value_axis": value_axis,
        },
    )

    return data, layout



import os

import pytest
import tempfile
import pandas as pd

import calliope
from calliope.test.common.util import check_error_or_warning


class TestModel:
    @pytest.fixture(scope="module")
    def national_scale_example(self):
        model = calliope.examples.national_scale(
            override_dict={"model.subset_time": "2005-01-01"}
        )
        model.run()
        return model

    def test_save_commented_model_yaml(self, national_scale_example):
        model = national_scale_example

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "model_debug.yaml")
            model.save_commented_model_yaml(out_path)
            assert os.path.isfile(out_path)

    def test_info(self, national_scale_example):
        model = national_scale_example

        model.info()

    def test_info_minimal_model(self):
        this_path = os.path.dirname(__file__)
        model_location = os.path.join(
            this_path, "common", "test_model", "model_minimal.yaml"
        )
        model = calliope.Model(model_location)

        model.info()

    def test_get_formatted_array_index(self, national_scale_example):
        array = national_scale_example.get_formatted_array(
            "resource", index_format="index"
        )

        assert array.dims == ("locs", "techs", "timesteps")

    def test_get_formatted_array_multiindex(self, national_scale_example):
        array = national_scale_example.get_formatted_array(
            "resource", index_format="multiindex"
        )

        assert array.dims == ("loc_techs_finite_resource", "timesteps")
        assert isinstance(array.loc_techs_finite_resource.to_index(), pd.MultiIndex)

    def test_get_formatted_array_unknown_format(self, national_scale_example):
        with pytest.raises(ValueError) as excinfo:
            national_scale_example.get_formatted_array("resource", index_format="foo")

        assert check_error_or_warning(
            excinfo, "Argument 'index_format' must be one of 'index' or 'multiindex'"
        )

    def test_get_formatted_array_unknown_var(self, national_scale_example):
        with pytest.raises(KeyError) as excinfo:
            national_scale_example.get_formatted_array("foo")

        assert check_error_or_warning(excinfo, "Variable foo not in Model data")



import logging

import xarray as xr
import pyomo.core as po  # pylint: disable=import-error

import calliope
from calliope.backend.pyomo.util import get_var, string_to_datetime
from calliope.backend import run as backend_run
from calliope.backend.pyomo import model as run_pyomo
import calliope.backend.pyomo.interface as pyomo_interface

from calliope.core.util.dataset import reorganise_xarray_dimensions
from calliope.core.util.logging import log_time
from calliope import exceptions
from calliope.core.attrdict import AttrDict
from calliope.postprocess.results import postprocess_model_results

logger = logging.getLogger(__name__)


def access_pyomo_model_inputs(backend_model):
    """
    If the user wishes to inspect the parameter values used as inputs in the backend
    model, they can access a new Dataset of all the backend model inputs, including
    defaults applied where the user did not specify anything for a loc::tech
    """

    all_params = {
        i.name: get_var(backend_model, i.name, sparse=True)
        for i in backend_model.component_objects()
        if isinstance(i, po.base.param.IndexedParam)
    }
    all_param_ds = reorganise_xarray_dimensions(xr.Dataset(all_params))
    all_param_ds = string_to_datetime(backend_model, all_param_ds)
    for var in all_param_ds.data_vars:
        all_param_ds[var].attrs["is_result"] = 0

    return all_param_ds


def update_pyomo_param(backend_model, opt, param, update_dict):
    """
    A Pyomo Param value can be updated without the user directly accessing the
    backend model.

    Parameters
    ----------
    param : str
        Name of the parameter to update
    update_dict : dict
        keys are parameter indeces (either strings or tuples of strings,
        depending on whether there is one or more than one dimension). Values
        are the new values being assigned to the parameter at the given indeces.

    Returns
    -------
    Value(s) will be updated in-place, requiring the user to run the model again to
    see the effect on results.

    """
    if not hasattr(backend_model, param):
        raise exceptions.ModelError(
            "Parameter `{}` not in the Pyomo Backend. Check that the string "
            "matches the corresponding constraint/cost in the model.inputs "
            "xarray Dataset".format(param)
        )
    elif not isinstance(getattr(backend_model, param), po.base.param.IndexedParam):
        raise exceptions.ModelError(
            "`{}` not a Parameter in the Pyomo Backend. Sets and decision variables "
            "cannot be updated by the user".format(param)
        )
    elif not isinstance(update_dict, dict):
        raise TypeError("`update_dict` must be a dictionary")
    else:
        getattr(backend_model, param).store_values(update_dict)

    if opt is not None and "persistent" in opt.name:
        exceptions.warn(
            "Updating the Pyomo parameter won't affect the optimisation run without also "
            "regenerating the relevant constraints or the objective function (see `regenerate_persistent_solver`)."
        )


def regenerate_persistent_pyomo_solver(backend_model, opt, constraints=None, obj=False):
    """
    Having updated a Pyomo Param or several of them, this function can be used
    to regenerate associated constraints in a persistent solver interface, such
    as "gurobi_persistent", before rerunning the model.
    The entire constraint need not be regenerated, it is possible to only point to
    those indexes whose associated parameters have changed.

    Parameters
    ----------
    constraints : dict of lists or None, default = None
        Names of constraints as keys and list of constraint index items as values,
        e.g. `{"energy_capacity_constraint": ["X1::pv"]}` or `{"cost_constraint": [("monetary", "X1::pv"), ("monetary", "X2::pv")])}`.
        Order of index values can be inferred by inspecting the constraint.
    obj : bool, default = False
        If True, will also regenerate the objective function.
    """
    if opt is None or "persistent" not in opt.name:
        raise exceptions.ModelError(
            "Can only regenerate persistent solvers. No persistent solver object found for this model run."
        )

    if obj is True:
        opt.set_objective(backend_model.obj)
    if constraints is not None:
        for constraint_name, constraint_idx in constraints.items():
            for idx in constraint_idx:
                opt.remove_constraint(getattr(backend_model, constraint_name)[idx])
                opt.add_constraint(getattr(backend_model, constraint_name)[idx])

    return opt


def activate_pyomo_constraint(backend_model, constraint, active=True):
    """
    Takes a constraint or objective name, finds it in the backend model and sets
    its status to either active or deactive.

    Parameters
    ----------
    constraint : str
        Name of the constraint/objective to activate/deactivate
        Built-in constraints include '_constraint'
    active: bool, default=True
        status to set the constraint/objective
    """
    if not hasattr(backend_model, constraint):
        raise exceptions.ModelError(
            "constraint/objective `{}` not in the Pyomo Backend.".format(constraint)
        )
    elif not isinstance(getattr(backend_model, constraint), po.base.Constraint):
        raise exceptions.ModelError(
            "`{}` not a constraint in the Pyomo Backend.".format(constraint)
        )
    elif active is True:
        getattr(backend_model, constraint).activate()
    elif active is False:
        getattr(backend_model, constraint).deactivate()
    else:
        raise ValueError("Argument `active` must be True or False")


def rerun_pyomo_model(model_data, backend_model, opt):
    """
    Rerun the Pyomo backend, perhaps after updating a parameter value,
    (de)activating a constraint/objective or updating run options in the model
    model_data object (e.g. `run.solver`).

    Returns
    -------
    new_model : calliope.Model
        New calliope model, including both inputs and results, but no backend interface.
    """
    backend_model.__calliope_run_config = AttrDict.from_yaml_string(
        model_data.attrs["run_config"]
    )
    timings = {}
    log_time(logger, timings, "model_creation")
    inputs = access_pyomo_model_inputs(backend_model)

    run_mode = backend_model.__calliope_run_config["mode"]
    if run_mode == "plan":
        kwargs = {}
    elif run_mode == "spores":
        kwargs = {"interface": pyomo_interface}
    else:
        raise exceptions.ModelError(
            "Cannot rerun the backend in {} run mode. Only `plan` or `spores` modes are "
            "possible.".format(run_mode)
        )
    run_func = getattr(backend_run, f"run_{run_mode}")
    results, backend_model, opt = run_func(
        model_data=model_data,
        timings=timings,
        backend=run_pyomo,
        build_only=False,
        backend_rerun=backend_model,
        opt=opt,
        **kwargs,
    )

    # Add additional post-processed result variables to results
    if results.attrs.get("termination_condition", None) in ["optimal", "feasible"]:
        results = postprocess_model_results(results, model_data, timings)

    for var in results.data_vars.values():
        var.attrs["is_result"] = 1

    for var in inputs.data_vars.values():
        var.attrs["is_result"] = 0

    new_model_data = xr.merge((results, inputs), compat="override")
    new_model_data.attrs.update(model_data.attrs)
    new_model_data.attrs.update(results.attrs)

    # Only add coordinates from the original model_data that don't already exist
    new_coords = [
        i for i in model_data.coords.keys() if i not in new_model_data.coords.keys()
    ]
    new_model_data = new_model_data.update(model_data[new_coords])

    # Reorganise the coordinates so that model data and new model data share
    # the same order of items in each dimension
    new_model_data = new_model_data.reindex(model_data.coords)

    exceptions.warn(
        "The results of rerunning the backend model are only available within "
        "the Calliope model returned by this function call."
    )

    new_calliope_model = calliope.Model(config=None, model_data=new_model_data)
    new_calliope_model._timings = timings

    return new_calliope_model


def add_pyomo_constraint(
    backend_model, constraint_name, constraint_sets, constraint_rule
):
    """
    Functionality to add a constraint to the Pyomo backend, by pointing it to
    a function which implements the mathematics

    Parameters
    ----------
    constraint_name : str
        Name for the pyomo.Constraint object
    constraint_sets : list of str
        Sets over which to implement the constraint.
        The list of sets in the Pyomo model corresponds to model dimensions and
        can be found in full by calling the backend interface method `get_all_model_attrs`
    constraint_rule : function with args (backend_model, *args)
        Function corresponding to the format expected by Pyomo,
        i.e. arguments include the Pyomo backend model (`backend_model`)
        and pointers to each set in `constraint_sets`.
        Set arguments need to be in the same order as in constraint_sets.
        The convention we follow is to name the index without the trailing `s`
        E.g. if 'timesteps' is in constraint_sets, then 'timestep' should be one of the function arguments;
        long `loc_techs` set names are abbreviated to `loc_tech`.
        To see what parameters and variables you have to work with,
        call the backend interface method `get_all_model_attrs`

    Examples
    --------
    To limit supply technologies' capacity to 90% of their maximum possible capacity,
    you would do the following:

    constraint_name = 'max_capacity_90_constraint'
    constraint_sets = ['loc_techs_supply']

    def max_capacity_90_constraint_rule(backend_model, loc_tech):

        return backend_model.energy_cap[loc_tech] <= (
            backend_model.energy_cap_max[loc_tech] * 0.9
        )

    # Add the constraint
    model.backend.add_constraint(constraint_name, constraint_sets, max_capacity_90_constraint_rule)

    # Rerun the model with new constraint.
    # Note: model.run(force_rerun=True) will *not* work, since the backend model will be rebuilt, killing any changes you've made.
    new_model = model.backend.rerun()

    Note that we like the convention that constraint names end with 'constraint' and
    constraint rules have the same text, with an appended '_rule',
    but you are not required to follow this convention to have a working constraint.
    """

    assert (
        constraint_rule.__code__.co_varnames[0] == "backend_model"
    ), "First argument of constraint function must be 'backend_model'."
    assert constraint_rule.__code__.co_argcount - 1 == len(
        constraint_sets
    ), "Number of constraint arguments must equal number of constraint sets + 1."

    try:
        sets = [getattr(backend_model, i) for i in constraint_sets]
    except AttributeError as e:
        e.args = (e.args[0].replace("'ConcreteModel'", "Pyomo backend model"),)
        raise

    setattr(
        backend_model,
        constraint_name,
        po.Constraint(*sets, **{"rule": constraint_rule}),
    )

    return backend_model


def get_all_pyomo_model_attrs(backend_model):
    """
    Get the name of all sets, parameters, and variables in the generated Pyomo model.

    Returns
    -------
    Dictionary differentiating between variables ('Var'), parameters ('Param'), and sets ('Set').
    variables and parameters are given as a dictionaries of lists, where keys are the item names and
    values are a list of dimensions over which they are indexed. These dimensions correspond to the
    sets.
    """
    # Indexed objected
    objects = {
        objname: {
            i.name: [j.name for j in i.index_set().subsets()]
            if i.name + "_index" == i.index_set().name
            else [i.index_set().name]
            for i in backend_model.component_objects()
            if isinstance(i, getattr(po.base, objname))
        }
        for objname in ["Var", "Param"]
    }
    # Indices
    objects["Set"] = [
        i.name for i in backend_model.component_objects() if isinstance(i, po.base.Set)
    ]

    return objects


class BackendInterfaceMethods:
    def __init__(self, model):
        self._backend = model._backend_model
        self._opt = model._backend_model_opt
        self._model_data = model._model_data

    def access_model_inputs(self):
        return access_pyomo_model_inputs(self._backend)

    access_model_inputs.__doc__ = access_pyomo_model_inputs.__doc__

    def update_param(self, *args, **kwargs):
        return update_pyomo_param(self._backend, self._opt, *args, **kwargs)

    update_param.__doc__ = update_pyomo_param.__doc__

    def activate_constraint(self, *args, **kwargs):
        return activate_pyomo_constraint(self._backend, *args, **kwargs)

    activate_constraint.__doc__ = activate_pyomo_constraint.__doc__

    def rerun(self, *args, **kwargs):
        return rerun_pyomo_model(
            self._model_data, self._backend, self._opt, *args, **kwargs
        )

    rerun.__doc__ = rerun_pyomo_model.__doc__

    def regenerate_persistent_solver(self, *args, **kwargs):
        self._opt = regenerate_persistent_pyomo_solver(
            self._backend, self._opt, *args, **kwargs
        )

    regenerate_persistent_solver.__doc__ = regenerate_persistent_pyomo_solver.__doc__

    def add_constraint(self, *args, **kwargs):
        self._backend = add_pyomo_constraint(self._backend, *args, **kwargs)

    add_constraint.__doc__ = add_pyomo_constraint.__doc__

    def get_all_model_attrs(self, *args, **kwargs):
        return get_all_pyomo_model_attrs(self._backend, *args, **kwargs)

    get_all_model_attrs.__doc__ = get_all_pyomo_model_attrs.__doc__
